import contextlib
from copy import copy
import os
import pickle
import datetime  # noqa: 251
from typing import Any, List, NamedTuple, Optional, Sequence
import humanize

from dlt.common import logger
from dlt.common.pendulum import pendulum
from dlt.common.configuration import is_secret_hint
from dlt.common.configuration.exceptions import ContextDefaultCannotBeCreated
from dlt.common.configuration.specs.config_section_context import ConfigSectionContext
from dlt.common.configuration.utils import get_resolved_traces
from dlt.common.configuration.container import Container
from dlt.common.exceptions import ExceptionTrace, ResourceNameNotAvailable
from dlt.common.logger import suppress_and_warn
from dlt.common.runtime.exec_info import TExecutionContext, get_execution_context
from dlt.common.pipeline import (
    ExtractInfo,
    LoadInfo,
    NormalizeInfo,
    PipelineContext,
    StepInfo,
    StepMetrics,
    SupportsPipeline,
)
from dlt.common.storages.file_storage import FileStorage
from dlt.common.typing import DictStrAny, StrAny, SupportsHumanize
from dlt.common.utils import uniq_id, get_exception_trace_chain

from dlt.extract.state import get_current_pipe_name

from dlt.pipeline.typing import TPipelineStep
from dlt.pipeline.exceptions import PipelineStepFailed


TRACE_ENGINE_VERSION = 1
TRACE_FILE_NAME = "trace.pickle"


# @dataclasses.dataclass(init=True)
class SerializableResolvedValueTrace(NamedTuple):
    """Information on resolved secret and config values"""

    key: str
    value: Any
    default_value: Any
    is_secret_hint: bool
    sections: Sequence[str]
    provider_name: str
    config_type_name: str

    def asdict(self) -> StrAny:
        """A dictionary representation that is safe to load."""
        return {k: v for k, v in self._asdict().items() if k not in ("value", "default_value")}

    def asstr(self, verbosity: int = 0) -> str:
        return f"{self.key}->{self.value} in {'.'.join(self.sections)} by {self.provider_name}"

    def __str__(self) -> str:
        return self.asstr(verbosity=0)


class _PipelineStepTrace(NamedTuple):
    span_id: str
    step: TPipelineStep
    started_at: datetime.datetime
    finished_at: datetime.datetime = None
    step_info: Optional[StepInfo[StepMetrics]] = None
    """A step outcome info ie. LoadInfo"""
    step_exception: Optional[str] = None
    """For failing steps contains exception string"""
    exception_traces: List[ExceptionTrace] = None
    """For failing steps contains traces of exception chain causing it"""


class PipelineStepTrace(SupportsHumanize, _PipelineStepTrace):
    """Trace of particular pipeline step, contains timing information, the step outcome info or exception in case of failing step with custom asdict()"""

    def asstr(self, verbosity: int = 0) -> str:
        completed_str = "FAILED" if self.step_exception else "COMPLETED"
        if self.started_at and self.finished_at:
            elapsed = self.finished_at - self.started_at
            elapsed_str = humanize.precisedelta(elapsed)
        else:
            elapsed_str = "---"
        msg = f"Step {self.step} {completed_str} in {elapsed_str}."
        if self.step_exception:
            msg += f"\nFailed due to: {self.step_exception}"
        if self.step_info and hasattr(self.step_info, "asstr"):
            info = self.step_info.asstr(verbosity)
            if info:
                msg += f"\n{info}"
        if verbosity > 0:
            msg += f"\nspan id: {self.span_id}"
        return msg

    def asdict(self) -> DictStrAny:
        """A dictionary representation of PipelineStepTrace that can be loaded with `dlt`"""
        d = self._asdict()
        if self.step_info:
            # name property depending on step name - generates nicer data
            d[f"{self.step}_info"] = step_info_dict = d.pop("step_info").asdict()
            d["step_info"] = {}
            # take only the base keys
            for prop in self.step_info._astuple()._asdict():
                if prop in step_info_dict:
                    d["step_info"][prop] = step_info_dict.pop(prop)
        # replace the attributes in exception traces with json dumps
        if self.exception_traces:
            # do not modify original traces
            d["exception_traces"] = copy(d["exception_traces"])
            traces: List[ExceptionTrace] = d["exception_traces"]
            for idx in range(len(traces)):
                if traces[idx].get("exception_attrs"):
                    # trace: ExceptionTrace
                    trace = traces[idx] = copy(traces[idx])
                    trace["exception_attrs"] = str(trace["exception_attrs"])  # type: ignore[typeddict-item]

        return d

    def __str__(self) -> str:
        return self.asstr(verbosity=0)


class _PipelineTrace(NamedTuple):
    """Pipeline runtime trace containing data on "extract", "normalize" and "load" steps and resolved config and secret values."""

    transaction_id: str
    pipeline_name: str
    execution_context: TExecutionContext
    started_at: datetime.datetime
    steps: List[PipelineStepTrace]
    """A list of steps in the trace"""
    finished_at: datetime.datetime = None
    resolved_config_values: List[SerializableResolvedValueTrace] = None
    """A list of resolved config values"""
    engine_version: int = TRACE_ENGINE_VERSION


class PipelineTrace(SupportsHumanize, _PipelineTrace):
    def asstr(self, verbosity: int = 0) -> str:
        last_step = self.steps[-1]
        completed_str = "FAILED" if last_step.step_exception else "COMPLETED"
        if self.started_at and self.finished_at:
            elapsed = self.finished_at - self.started_at
            elapsed_str = humanize.precisedelta(elapsed)
        else:
            elapsed_str = "---"
        msg = (
            f"Run started at {self.started_at} and {completed_str} in {elapsed_str} with"
            f" {len(self.steps)} steps."
        )
        if verbosity > 0 and len(self.resolved_config_values) > 0:
            msg += "\nFollowing config and secret values were resolved:\n"
            msg += "\n".join([s.asstr(verbosity) for s in self.resolved_config_values])
            msg += "\n"
        if len(self.steps) > 0:
            msg += "\n" + "\n\n".join([s.asstr(verbosity) for s in self.steps])
        return msg

    def last_pipeline_step_trace(self, step_name: TPipelineStep) -> PipelineStepTrace:
        matching_steps = [step for step in self.steps if step.step == step_name]
        if not matching_steps:
            return None
        return max(matching_steps, key=lambda step: step.started_at)

    def asdict(self) -> DictStrAny:
        """A dictionary representation of PipelineTrace that can be loaded with `dlt`"""
        d = self._asdict()
        # run step is the same as load step
        d["steps"] = [step.asdict() for step in self.steps if step.step != "run"]
        return d

    @property
    def last_extract_info(self) -> ExtractInfo:
        step_trace = self.last_pipeline_step_trace("extract")
        if step_trace and isinstance(step_trace.step_info, ExtractInfo):
            return step_trace.step_info
        return None

    @property
    def last_normalize_info(self) -> NormalizeInfo:
        step_trace = self.last_pipeline_step_trace("normalize")
        if step_trace and isinstance(step_trace.step_info, NormalizeInfo):
            return step_trace.step_info
        return None

    @property
    def last_load_info(self) -> LoadInfo:
        step_trace = self.last_pipeline_step_trace("load")
        if step_trace and isinstance(step_trace.step_info, LoadInfo):
            return step_trace.step_info
        return None

    def __str__(self) -> str:
        return self.asstr(verbosity=0)


from dlt.common.runtime.tracking import SupportsTracking


# plug in your own tracking modules here
TRACKING_MODULES: List[SupportsTracking] = None


def start_trace(step: TPipelineStep, pipeline: SupportsPipeline) -> PipelineTrace:
    trace = PipelineTrace(
        uniq_id(),
        pipeline.pipeline_name,
        get_execution_context(),
        pendulum.now(),
        steps=[],
        resolved_config_values=[],
    )
    for module in TRACKING_MODULES:
        with suppress_and_warn(f"on_start_trace on module {module} failed"):
            module.on_start_trace(trace, step, pipeline)
    with suppress_and_warn(
        f"on_start_trace on collector `{pipeline.collector.__class__.__name__}` failed"
    ):
        pipeline.collector.on_start_trace(trace, step, pipeline)

    return trace


def start_trace_step(
    trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline
) -> PipelineStepTrace:
    trace_step = PipelineStepTrace(uniq_id(), step, pendulum.now())
    for module in TRACKING_MODULES:
        with suppress_and_warn(f"start_trace_step on module {module} failed"):
            module.on_start_trace_step(trace, step, pipeline)
    with suppress_and_warn(
        f"on_start_trace_step on collector `{pipeline.collector.__class__.__name__}` failed"
    ):
        pipeline.collector.on_start_trace_step(trace, step, pipeline)
    return trace_step


def end_trace_step(
    trace: PipelineTrace,
    step: PipelineStepTrace,
    pipeline: SupportsPipeline,
    step_info: Any,
    send_state: bool,
) -> PipelineTrace:
    # saves runtime trace of the pipeline
    if isinstance(step_info, PipelineStepFailed):
        exception_traces = get_exception_traces(step_info)
        step_exception = str(step_info)
        step_info = step_info.step_info
    elif isinstance(step_info, Exception):
        exception_traces = get_exception_traces(step_info)
        step_exception = str(step_info)
        if step_info.__context__:
            step_exception += "caused by: " + str(step_info.__context__)
        step_info = None
    else:
        exception_traces = None
        step_exception = None

    step = step._replace(
        finished_at=pendulum.now(),
        step_exception=step_exception,
        exception_traces=exception_traces,
        step_info=step_info,
    )

    # this will collect traces from the last clear (happens at the end of trace)
    resolved_values = map(
        lambda v: SerializableResolvedValueTrace(
            v.key,
            None if is_secret_hint(v.hint) else v.value,
            None if is_secret_hint(v.hint) else v.default_value,
            is_secret_hint(v.hint),
            v.sections,
            v.provider_name,
            str(type(v.config).__qualname__),
        ),
        get_resolved_traces().resolved_traces,
    )
    trace.resolved_config_values[:] = list(resolved_values)
    trace.steps.append(step)
    for module in TRACKING_MODULES:
        with suppress_and_warn(f"end_trace_step on module {module} failed"):
            module.on_end_trace_step(trace, step, pipeline, step_info, send_state)
    with suppress_and_warn(
        f"on_end_trace_step on collector `{pipeline.collector.__class__.__name__}` failed"
    ):
        pipeline.collector.on_end_trace_step(trace, step, pipeline, step_info, send_state)
    return trace


def end_trace(trace: PipelineTrace, pipeline: SupportsPipeline, send_state: bool) -> PipelineTrace:
    trace = trace._replace(finished_at=pendulum.now())
    for module in TRACKING_MODULES:
        with suppress_and_warn(f"end_trace on module {module} failed"):
            module.on_end_trace(trace, pipeline, send_state)
    with suppress_and_warn(
        f"on_end_trace on collector `{pipeline.collector.__class__.__name__}` failed"
    ):
        pipeline.collector.on_end_trace(trace, pipeline, send_state)
    # clear collected config resolver traces
    get_resolved_traces().clear()
    return trace


def merge_traces(last_trace: PipelineTrace, new_trace: PipelineTrace) -> PipelineTrace:
    """Merges `new_trace` into `last_trace` by combining steps and timestamps. `new_trace` replace the `last_trace` if it has more than 1 step.`"""
    if len(new_trace.steps) > 1 or last_trace is None:
        return new_trace

    last_trace.steps.extend(new_trace.steps)
    # remember only last 100 steps and keep the finished up from previous trace
    return last_trace._replace(
        steps=last_trace.steps[-100:],
        finished_at=new_trace.finished_at,
        resolved_config_values=new_trace.resolved_config_values,
    )


def save_trace(trace_dir: str, trace: PipelineTrace) -> None:
    # remove previous file, we do not want to keep old trace even if we fail later
    trace_dump_path = os.path.join(trace_dir, TRACE_FILE_NAME)
    if os.path.isfile(trace_dump_path):
        os.unlink(trace_dump_path)
    with suppress_and_warn("Failed to create trace dump via pickle"):
        trace_dump = pickle.dumps(trace)
        FileStorage.save_atomic(trace_dir, TRACE_FILE_NAME, trace_dump, file_type="b")


def load_trace(trace_dir: str, ignore_errors: bool = True) -> PipelineTrace:
    try:
        with open(os.path.join(trace_dir, TRACE_FILE_NAME), mode="rb") as f:
            return pickle.load(f)  # type: ignore
    except FileNotFoundError:
        # file not found return no trace
        pass
    except Exception as ex:
        # we do not guarantee traces backward compat
        if not ignore_errors:
            raise
        logger.warning(f"Error when loading trace at {trace_dir}: %s" % ex)

    return None


def get_trace_file_path(pipelines_dir: str, pipeline_name: str) -> str:
    """Get the path to the pickle file for a pipeline"""
    return os.path.join(pipelines_dir, pipeline_name, TRACE_FILE_NAME)


def get_exception_traces(exc: BaseException, container: Container = None) -> List[ExceptionTrace]:
    """Gets exception trace chain and extend it with data available in Container context"""
    traces = get_exception_trace_chain(exc)
    container = container or Container()

    # get resource name
    resource_name: str = None
    with contextlib.suppress(ResourceNameNotAvailable):
        resource_name = get_current_pipe_name()
    # get source name
    source_name: str = None
    with contextlib.suppress(ContextDefaultCannotBeCreated):
        sections_context = container[ConfigSectionContext]
        source_name = sections_context.source_state_key
    # get pipeline name
    proxy = container[PipelineContext]
    if proxy.is_active():
        pipeline_name = proxy.pipeline().pipeline_name
    else:
        pipeline_name = None

    # apply context to trace
    for trace in traces:
        # only to dlt exceptions
        if "exception_attrs" in trace:
            trace.setdefault("resource_name", resource_name)
            trace.setdefault("pipeline_name", pipeline_name)
            trace.setdefault("source_name", source_name)

    return traces



from dataclasses import dataclass
import os
import pathlib
import posixpath
from typing import Any, Literal, Optional, Type, ClassVar, Dict, Union
from urllib.parse import urlparse, unquote, urlunparse

from dlt.common.configuration import configspec, resolve_type
from dlt.common.configuration.exceptions import ConfigurationValueError
from dlt.common.configuration.specs import CredentialsConfiguration
from dlt.common.configuration.specs import (
    GcpServiceAccountCredentials,
    AwsCredentials,
    GcpOAuthCredentials,
    AnyAzureCredentials,
    HfCredentials,
    BaseConfiguration,
    SFTPCredentials,
)
from dlt.common.configuration.specs.base_configuration import NotResolved
from dlt.common.exceptions import TerminalValueError
from dlt.common.typing import Annotated, DictStrAny, DictStrOptionalStr, get_args
from dlt.common.utils import digest128


TSchemaFileFormat = Literal["json", "yaml", "dbml", "dot", "mermaid"]
SCHEMA_FILES_EXTENSIONS = get_args(TSchemaFileFormat)


@configspec
class SchemaStorageConfiguration(BaseConfiguration):
    schema_volume_path: str = None  # path to volume with default schemas
    import_schema_path: Optional[str] = None  # path from which to import a schema into storage
    export_schema_path: Optional[str] = None  # path to which export schema from storage
    external_schema_format: TSchemaFileFormat = "yaml"  # format in which to expect external schema
    external_schema_format_remove_defaults: bool = (
        True  # remove default values when exporting schema
    )


@configspec
class NormalizeStorageConfiguration(BaseConfiguration):
    normalize_volume_path: str = None  # path to volume where normalized loader files will be stored


@configspec
class LoadStorageConfiguration(BaseConfiguration):
    load_volume_path: str = (
        None  # path to volume where files to be loaded to analytical storage are stored
    )
    delete_completed_jobs: bool = (
        False  # if set to true the folder with completed jobs will be deleted
    )


FileSystemCredentials = Union[
    AwsCredentials,
    GcpServiceAccountCredentials,
    AnyAzureCredentials,
    GcpOAuthCredentials,
    SFTPCredentials,
    HfCredentials,
]


def ensure_canonical_az_url(
    bucket_url: str, target_scheme: str, storage_account_name: str = None, account_host: str = None
) -> str:
    """Converts any of the forms of azure blob storage into canonical form of {target_scheme}://<container_name>@<storage_account_name>.{account_host}/path

    `azure_storage_account_name` is optional only if not present in bucket_url, `account_host` assumes "<azure_storage_account_name>.dfs.core.windows.net" by default
    """
    parsed_bucket_url = urlparse(bucket_url)
    # Converts an az://<container_name>/<path> to abfss://<container_name>@<storage_account_name>.dfs.core.windows.net/<path>
    if parsed_bucket_url.username:
        # has the right form, ensure abfss schema
        return urlunparse(parsed_bucket_url._replace(scheme=target_scheme))

    if not storage_account_name and not account_host:
        raise TerminalValueError(
            f"Could not convert azure blob storage url `{bucket_url}` into canonical form "
            f" ({target_scheme}://<container_name>@<storage_account_name>.dfs.core.windows.net/<path>)"
            f" because storage account name is not known. Please use {target_scheme}:// canonical"
            " url as bucket_url in filesystem credentials"
        )

    account_host = account_host or f"{storage_account_name}.dfs.core.windows.net"
    netloc = (
        f"{parsed_bucket_url.netloc}@{account_host}" if parsed_bucket_url.netloc else account_host
    )

    # as required by databricks
    _path = parsed_bucket_url.path
    return urlunparse(
        parsed_bucket_url._replace(
            scheme=target_scheme,
            netloc=netloc,
            path=_path,
        )
    )


def _make_sftp_url(scheme: str, fs_path: str, bucket_url: str) -> str:
    parsed_bucket_url = urlparse(bucket_url)
    return f"{scheme}://{parsed_bucket_url.hostname}{fs_path}"


def _make_az_url(scheme: str, fs_path: str, bucket_url: str) -> str:
    parsed_bucket_url = urlparse(bucket_url)
    if parsed_bucket_url.username:
        # az://<container_name>@<storage_account_name>.dfs.core.windows.net/<path>
        # fs_path always starts with container
        split_path = fs_path.split("/", maxsplit=1)
        # preserve slash at the end
        if len(split_path) == 2 and split_path[1] == "":
            split_path[1] = "/"
        # if just a container name, add empty path
        if len(split_path) == 1:
            split_path.append("")
        container, path = split_path
        netloc = f"{container}@{parsed_bucket_url.hostname}"
        # this strips trailing slash
        uri = urlunparse(parsed_bucket_url._replace(path=path, scheme=scheme, netloc=netloc))
        return uri
    return f"{scheme}://{fs_path}"


def _make_file_url(scheme: str, fs_path: str, bucket_url: str) -> str:
    """Creates a normalized file:// url from a local path

    netloc is never set. UNC paths are represented as file://host/path
    """
    p_ = pathlib.Path(fs_path)
    # will remove trailing separator
    p_ = p_.expanduser().resolve()
    uri = p_.as_uri()
    if fs_path.endswith(os.path.sep):
        uri += "/"
    return uri


def _make_http_url(schema: str, fs_path: str, bucket_url: str) -> str:
    parsed_http_url = urlparse(fs_path, schema)
    return urlunparse(parsed_http_url)


MAKE_URI_DISPATCH = {
    "az": _make_az_url,
    "file": _make_file_url,
    "sftp": _make_sftp_url,
    "https": _make_http_url,
}

MAKE_URI_DISPATCH["adl"] = MAKE_URI_DISPATCH["az"]
MAKE_URI_DISPATCH["abfs"] = MAKE_URI_DISPATCH["az"]
MAKE_URI_DISPATCH["azure"] = MAKE_URI_DISPATCH["az"]
MAKE_URI_DISPATCH["abfss"] = MAKE_URI_DISPATCH["az"]
MAKE_URI_DISPATCH["local"] = MAKE_URI_DISPATCH["file"]


def make_fsspec_url(scheme: str, fs_path: str, bucket_url: str) -> str:
    """Creates url from `fs_path` and `scheme` using bucket_url as an `url` template, if `fs_path`
    ends with separator (indicating folder), it is preserved

    Args:
        scheme (str): scheme of the resulting url
        fs_path (str): kind of absolute path that fsspec uses to locate resources for particular filesystem.
        bucket_url (str): an url template. the structure of url will be preserved if possible
    """
    _maker = MAKE_URI_DISPATCH.get(scheme)
    if _maker:
        return _maker(scheme, fs_path, bucket_url)
    return f"{scheme}://{fs_path}"


@configspec
class FilesystemConfiguration(BaseConfiguration):
    """A configuration defining filesystem location and access credentials.

    When configuration is resolved, `bucket_url` is used to extract a protocol and request corresponding credentials class.
    * s3 (AwsCredentials)
    * gs, gcs (GcpServiceAccountCredentials, GcpOAuthCredentials)
    * az, abfs, adl, abfss, azure (AzureCredentials, AzureServicePrincipalCredentials)
    * file, memory
    * gdrive (GcpServiceAccountCredentials, GcpOAuthCredentials)
    * hf (HfCredentials)
    * sftp (SFTPCredentials)
    """

    PROTOCOL_CREDENTIALS: ClassVar[Dict[str, Any]] = {
        "gs": Union[GcpServiceAccountCredentials, GcpOAuthCredentials],
        "gcs": Union[GcpServiceAccountCredentials, GcpOAuthCredentials],
        "gdrive": Union[GcpServiceAccountCredentials, GcpOAuthCredentials],
        "s3": AwsCredentials,
        "az": AnyAzureCredentials,
        "abfs": AnyAzureCredentials,
        "adl": AnyAzureCredentials,
        "abfss": AnyAzureCredentials,
        "azure": AnyAzureCredentials,
        "hf": HfCredentials,
        "sftp": SFTPCredentials,
    }

    bucket_url: str = None

    # should be a union of all possible credentials as found in PROTOCOL_CREDENTIALS
    credentials: FileSystemCredentials = None

    read_only: bool = False
    """Indicates read only filesystem access. Will enable caching"""
    kwargs: Optional[DictStrAny] = None
    """Additional arguments passed to fsspec constructor ie. dict(use_ssl=True) for s3fs"""
    client_kwargs: Optional[DictStrAny] = None
    """Additional arguments passed to underlying fsspec native client ie. dict(verify="public.crt) for botocore"""
    config_kwargs: Optional[DictStrAny] = None
    """Additional arguments as Config in botocore"""

    @property
    def protocol(self) -> str:
        """`bucket_url` protocol"""
        return self.parse_protocol(self.bucket_url)

    @property
    def is_local_filesystem(self) -> bool:
        return self.protocol == "file"

    @property
    def pathlib(self) -> Any:
        """Returns pathlib suitable for joining and other path ops"""
        return os.path if self.is_local_path(self.bucket_url) else posixpath

    def on_resolved(self) -> None:
        self.verify_bucket_url()

    def on_partial(self) -> None:
        if self.bucket_url:
            self.verify_bucket_url()

    def normalize_bucket_url(self) -> None:
        """Normalizes bucket_url ie. by making local paths absolute and converting to file:"""
        # save original url
        self._orig_bucket_url = self.bucket_url
        # this is just a path in a local file system
        if self.is_local_path(self.bucket_url):
            self.bucket_url = self.make_file_url(self.bucket_url)

    def verify_bucket_url(self) -> None:
        url = urlparse(self.bucket_url)
        if not url.path and not url.netloc:
            raise ConfigurationValueError(
                "File `path` and `netloc` are missing. Field `bucket_url` of"
                " `FilesystemClientConfiguration` must contain valid url with a path or"
                " host:password component."
            )
        self.normalize_bucket_url()

    def original_bucket_url(self) -> str:
        """Returns bucket_url before normalization"""
        return self._orig_bucket_url

    @resolve_type("credentials")
    def resolve_credentials_type(self) -> Type[CredentialsConfiguration]:
        # use known credentials or empty credentials for unknown protocol
        return self.PROTOCOL_CREDENTIALS.get(self.protocol) or Optional[CredentialsConfiguration]  # type: ignore[return-value]

    def fingerprint(self) -> str:
        """Returns a fingerprint of bucket schema and netloc.

        Returns:
            str: Fingerprint.
        """
        if not self.bucket_url:
            return ""

        if self.is_local_path(self.bucket_url):
            return digest128("")

        url = urlparse(self.bucket_url)
        return digest128(self.bucket_url.replace(url.path, ""))

    def make_url(self, fs_path: str) -> str:
        """Makes a full url (with scheme) form fs_path which is kind-of absolute path used by fsspec to identify resources.
        This method will use `bucket_url` to infer the original form of the url.
        """
        return make_fsspec_url(self.protocol, fs_path, self.bucket_url)

    def __str__(self) -> str:
        """Return displayable destination location"""
        url = urlparse(self.bucket_url)
        # do not show passwords
        if url.password:
            new_netloc = f"{url.username}:****@{url.hostname}"
            if url.port:
                new_netloc += f":{url.port}"
            return url._replace(netloc=new_netloc).geturl()
        return self.bucket_url

    @staticmethod
    def is_local_path(url: str) -> bool:
        """Checks if `url` is a local path, without a schema"""
        url_parsed = urlparse(url)
        # this prevents windows absolute paths to be recognized as schemas
        return not url_parsed.scheme or os.path.isabs(url)

    @staticmethod
    def make_local_path(file_url: str) -> str:
        """Gets a valid local filesystem path from file:// scheme.
        Supports POSIX/Windows/UNC paths

        Returns:
            str: local filesystem path
        """
        url = urlparse(file_url)
        if url.scheme != "file":
            raise ValueError(f"Must be file scheme but is `{url.scheme}`")
        if not url.path and not url.netloc:
            raise ConfigurationValueError("File `path` and `netloc` are missing.")
        local_path = unquote(url.path)
        if url.netloc:
            # or UNC file://localhost/path
            local_path = "//" + unquote(url.netloc) + local_path
        else:
            # if we are on windows, strip the POSIX root from path which is always absolute
            if os.path.sep != local_path[0]:
                # filesystem root
                if local_path == "/":
                    return str(pathlib.Path("/").resolve())
                # this prevents /C:/ or ///share/ where both POSIX and Windows root are present
                if os.path.isabs(local_path[1:]):
                    local_path = local_path[1:]
        return str(pathlib.Path(local_path))

    @staticmethod
    def make_file_url(local_path: str) -> str:
        """Creates a normalized file:// url from a local path

        netloc is never set. UNC paths are represented as file://host/path
        """
        return make_fsspec_url("file", local_path, None)

    @staticmethod
    def parse_protocol(bucket_url: str) -> str:
        """Parses protocol from bucket_url.

        Returns "file" for local paths without scheme.
        """

        if FilesystemConfiguration.is_local_path(bucket_url):
            return "file"
        else:
            return urlparse(bucket_url).scheme


@dataclass
class WithLocalFiles(BaseConfiguration):
    """Mixin to BaseConfiguration that shifts relative locations into `local_dir` and allows for a few special locations.
    :pipeline: in the pipeline working folder
    :external: location is an instance of an object (ie. fsspec or duckdb connection)

    Unifies default names for objects stored in filesystem (datasets / database names)
    - if destination_name is known it will be used to create name OR
    - if pipeline_name is known it will be used to create name

    Note: if not provided explicitly, `local_dir` is taken from run context and destination_name is injected
    Pipeline class which instantiates configuration will bind all NotResolved() params below explicitly

    """

    destination_type: Optional[str] = None
    destination_name: Optional[str] = None

    local_dir: Annotated[str, NotResolved()] = None
    pipeline_name: Annotated[Optional[str], NotResolved()] = None
    # needed by duckdb
    # TODO: deprecate this in 2.0
    pipeline_working_dir: Annotated[Optional[str], NotResolved()] = None
    legacy_db_path: Annotated[Optional[str], NotResolved()] = None

    def on_partial(self) -> None:
        from dlt.common.runtime.run_context import active

        if not self.local_dir:
            self.local_dir = os.path.abspath(active().local_dir)
            if not self.is_partial():
                self.resolve()

    def attach_from(self, other: "WithLocalFiles") -> None:
        """Attaches to the same local files info as `other`"""
        self.local_dir = other.local_dir
        self.destination_name = other.destination_name
        self.pipeline_name = other.pipeline_name
        self.pipeline_working_dir = other.pipeline_working_dir
        self.legacy_db_path = other.legacy_db_path

    def make_default_location(self, default_location_pat: str) -> str:
        # use destination name to create default location name
        if self.destination_name:
            return default_location_pat % self.destination_name
        else:
            return default_location_pat % (self.pipeline_name or self.destination_type)

    def make_location(self, configured_location: str, default_location_pat: str) -> str:
        # do not set any paths for external database / instance
        if configured_location == ":external:":
            return configured_location

        if configured_location == ":pipeline:":
            # try the pipeline context
            if self.pipeline_working_dir:
                return os.path.join(
                    self.pipeline_working_dir, self.make_default_location(default_location_pat)
                )
            raise RuntimeError(
                "Attempting to use special location `:pipeline:` outside of pipeline context."
            )
        else:
            # if explicit path is absolute, use it
            if configured_location and os.path.isabs(configured_location):
                return configured_location
            # TODO: restore this check for the paths below. we may require that path exists
            #  if pipeline had already run
            # if not self.bound_to_pipeline.first_run:
            #     if not os.path.exists(pipeline_path):
            #         logger.warning(
            #             f"Duckdb attached to pipeline {self.bound_to_pipeline.pipeline_name} in"
            #             f" path {os.path.relpath(pipeline_path)} was could not be found but"
            #             " pipeline has already ran. This may be a result of (1) recreating or"
            #             " attaching pipeline  without or with changed explicit path to database"
            #             " that was used when creating the pipeline. (2) keeping the path to to"
            #             " database in secrets and changing the current working folder so  dlt"
            #             " cannot see them. (3) you deleting the database."
            # use stored path if relpath
            if self.legacy_db_path:
                return self.legacy_db_path
            # use tmp path as root, not cwd
            return os.path.join(
                self.local_dir,
                configured_location or self.make_default_location(default_location_pat),
            )


@configspec
class FilesystemConfigurationWithLocalFiles(FilesystemConfiguration, WithLocalFiles):
    """FilesystemConfiguration subclass that adjust relative local filesystem bucket_url to
    be relative to `local_dir`.
    """

    def normalize_bucket_url(self) -> None:
        # here we deal with normalized file:// local paths
        if self.is_local_filesystem:
            # convert to native path
            try:
                local_file_path = self.make_local_path(self.bucket_url)
            except ValueError:
                local_file_path = self.bucket_url
            relocated_path = self.make_location(local_file_path, "%s")
            # convert back into file:// schema if relocated
            if local_file_path != relocated_path:
                if self.bucket_url.startswith("file:"):
                    self.bucket_url = self.make_file_url(relocated_path)
                else:
                    self.bucket_url = relocated_path
        # modified local path before it is normalized
        super().normalize_bucket_url()

    @resolve_type("credentials")
    def resolve_credentials_type(self) -> Type[CredentialsConfiguration]:
        return super().resolve_credentials_type()

    @staticmethod
    def make_file_url(local_path: str) -> str:
        """Creates a normalized file:// url from a local path. If path is relative, `local_dir`
        will be used.
        netloc is never set. UNC paths are represented as file://host/path
        """
        if not os.path.isabs(local_path):
            from dlt.common.runtime.run_context import active

            local_path = os.path.join(active().local_dir, local_path)

        return make_fsspec_url("file", local_path, None)



import abc
from typing import Any, Dict, Generic, Type, Generator, Protocol, Tuple, TYPE_CHECKING, TypeVar

from dlt.common.typing import DictStrAny, TDataItem, StrAny
from dlt.common.data_types.typing import TDataType

if TYPE_CHECKING:
    from dlt.common.schema import Schema
else:
    Schema = Any

# type definitions for json normalization function

# iterator of form ((table_name, parent_path, ident_path), dict) must be returned from normalization function
TNormalizedRowIterator = Generator[
    Tuple[Tuple[str, Tuple[str, ...], Tuple[str, ...]], StrAny], bool, None
]

# type var for data item normalizer config
TNormalizerConfig = TypeVar("TNormalizerConfig", bound=Any)


class DataItemNormalizer(abc.ABC, Generic[TNormalizerConfig]):
    @abc.abstractmethod
    def __init__(self, schema: Schema) -> None:
        pass

    @abc.abstractmethod
    def normalize_data_item(
        self, item: TDataItem, load_id: str, table_name: str
    ) -> TNormalizedRowIterator:
        pass

    @abc.abstractmethod
    def extend_schema(self, extend_tables: bool = True) -> None:
        pass

    @abc.abstractmethod
    def extend_table(self, table_name: str) -> None:
        pass

    @abc.abstractmethod
    def remove_table(self, table_name: str) -> None:
        pass

    @property
    def py_type_to_sc_type_map(self) -> Dict[Type[Any], TDataType]:
        """Lookup from Python type to schema type. Override in subclasses."""
        return {}

    @abc.abstractmethod
    def py_type_to_sc_type(self, t: Type[Any]) -> TDataType:
        """Resolve Python type to schema type including subclass/Enum handling."""
        ...

    @abc.abstractmethod
    def can_coerce_type(self, to_type: TDataType, from_type: TDataType) -> bool:
        """Return True if `from_type` can be coerced to `to_type`."""
        ...

    @abc.abstractmethod
    def coerce_type(self, to_type: TDataType, from_type: TDataType, value: Any) -> Any:
        """Coerce `value` from `from_type` to `to_type`. Raises ValueError on failure."""
        ...

    @classmethod
    @abc.abstractmethod
    def update_normalizer_config(cls, schema: Schema, config: TNormalizerConfig) -> None:
        pass

    @classmethod
    @abc.abstractmethod
    def get_normalizer_config(cls, schema: Schema) -> TNormalizerConfig:
        pass


class SupportsDataItemNormalizer(Protocol):
    """Expected of modules defining data item normalizer"""

    DataItemNormalizer: Type[DataItemNormalizer[Any]]
    """A class with a name DataItemNormalizer deriving from normalizers.json.DataItemNormalizer"""


def wrap_in_dict(label: str, item: Any) -> DictStrAny:
    """Wraps `item` that is not a dictionary into dictionary that can be json normalized"""
    return {label: item}


__all__ = [
    "TNormalizedRowIterator",
    "TNormalizerConfig",
    "DataItemNormalizer",
    "SupportsDataItemNormalizer",
    "wrap_in_dict",
]



from dlt.common.normalizers.naming.snake_case import NamingConvention as SnakeCaseNamingConvention


S3_TABLES_MAX_IDENTIFIER_LENGTH = 255


class NamingConvention(SnakeCaseNamingConvention):
    """Naming convention for S3 Tables.

    Built to comply with the rules listed here: https://docs.aws.amazon.com/AmazonS3/latest/userguide/s3-tables-buckets-naming.html#naming-rules-table.

    Extends `snake_case` naming convention with additions:
    - enforce max identifier length of 255 characters
    - remove leading underscores from table identifiers

    Note that this S3 Tables rule is currently not checked/enforced by the naming convention:
    - "Namespace names must not start with the reserved prefix `aws`."
    """

    def __init__(self, max_length: int = None) -> None:
        super().__init__(max_length)
        if self.max_length and self.max_length > S3_TABLES_MAX_IDENTIFIER_LENGTH:
            raise ValueError(
                "`max_length` for `s3_tables` naming convention may not exceed"
                f" {S3_TABLES_MAX_IDENTIFIER_LENGTH}"
            )
        elif self.max_length is None:
            self.max_length = S3_TABLES_MAX_IDENTIFIER_LENGTH

    def normalize_table_identifier(self, identifier: str) -> str:
        norm_identifier = super().normalize_table_identifier(identifier)
        norm_identifier = self._remove_leading_underscores(norm_identifier)
        return norm_identifier



import warnings
from dlt.common.warnings import Dlt100DeprecationWarning
from dlt.common.destination.configuration import (
    CsvQuoting,
    CsvFormatConfiguration,
    ParquetFormatConfiguration,
)

warnings.warn(
    "Please import format configuration from dlt.common.destination.configuration",
    Dlt100DeprecationWarning,
    stacklevel=2,
)

__all__ = ["CsvQuoting", "CsvFormatConfiguration", "ParquetFormatConfiguration"]



import sys
from types import ModuleType
from typing import Dict, Any, Optional

from dlt.common.destination.client import SupportsOpenTables
from dlt.common.pipeline import SupportsPipeline
from dlt.common.schema.typing import TTableFormat


def load_open_tables(
    pipeline: SupportsPipeline,
    table_format: TTableFormat,
    *tables: str,
    schema_name: Optional[str] = None,
    include_dlt_tables: bool = False,
) -> Dict[str, Any]:
    with pipeline.destination_client(schema_name=schema_name) as client:
        assert isinstance(
            client, SupportsOpenTables
        ), "This requires destination that supports open tables via SupportsOpenTables interface."

        all_tables = (
            client.schema.tables.values() if include_dlt_tables else client.schema.data_tables()
        )

        schema_open_tables = [
            t["name"] for t in all_tables if client.is_open_table(table_format, t["name"])
        ]
        if len(tables) > 0:
            invalid_tables = set(tables) - set(schema_open_tables)
            if len(invalid_tables) > 0:
                available_schemas_info = ""
                available_schemas = list(pipeline.schemas.keys())
                if len(available_schemas) > 1:
                    available_schemas_info = f" Available schemas are {available_schemas}"
                raise ValueError(
                    f"Schema `{client.schema.name}` does not contain `{table_format=:}` tables with"
                    f" names: {', '.join(invalid_tables)}.{available_schemas_info}"
                )
            schema_open_tables = [t for t in schema_open_tables if t in tables]

        return {name: client.load_open_table(table_format, name) for name in schema_open_tables}


def is_instance_lib(obj: Any, *, class_ref: str) -> bool:
    """Allows `isinstance()` checks without directly importing 3rd party libraries

    Example:
        ```python
        df = pd.DataFrame(...)
        _isinstance_external(df, class_ref="pandas.DataFrame")
        ```
    """
    import_parts = class_ref.split(".")
    module_name = import_parts[0]

    if module_name not in sys.modules:
        return False

    module: ModuleType = sys.modules[module_name]
    target_class: Any = module
    for part in import_parts[1:]:
        target_class = getattr(target_class, part)

    return isinstance(obj, target_class)



from dlt.common.data_types.type_helpers import coerce_value, py_type_to_sc_type
from dlt.common.data_types.typing import TDataType, DATA_TYPES

__all__ = ["coerce_value", "py_type_to_sc_type", "TDataType", "DATA_TYPES"]



from copy import copy
from typing import Optional, Dict, Any, Union

from dlt.common.pendulum import pendulum
from dlt.common.exceptions import MissingDependencyException
from dlt.common.typing import TSecretStrValue, Self
from dlt.common.configuration.specs import (
    CredentialsConfiguration,
    CredentialsWithDefault,
    configspec,
)
from dlt.common.configuration.specs.mixins import WithObjectStoreRsCredentials, WithPyicebergConfig
from dlt import version
from dlt.common.utils import without_none

_AZURE_STORAGE_EXTRA = f"{version.DLT_PKG_NAME}[az]"


@configspec
class AzureCredentialsBase(CredentialsConfiguration, WithObjectStoreRsCredentials):
    azure_storage_account_name: str = None
    azure_account_host: Optional[str] = None
    """Alternative host when accessing blob storage endpoint ie. my_account.dfs.core.windows.net"""

    def to_adlfs_credentials(self) -> Dict[str, Any]:
        pass

    def to_object_store_rs_credentials(self) -> Dict[str, str]:
        # https://docs.rs/object_store/latest/object_store/azure
        creds: Dict[str, Any] = without_none(self.to_adlfs_credentials())  # type: ignore[assignment]
        # only string options accepted
        creds.pop("anon", None)

        if isinstance(self, CredentialsWithDefault) and self.has_default_credentials():
            dc = self.default_credentials()

            # https://learn.microsoft.com/en-us/azure/storage/blobs/authorize-access-azure-active-directory#microsoft-authentication-library-msal
            creds["azure_storage_token"] = dc.get_token("https://storage.azure.com/.default").token

            return creds

        return creds


@configspec
class AzureCredentialsWithoutDefaults(AzureCredentialsBase, WithPyicebergConfig):
    """Credentials for Azure Blob Storage, compatible with adlfs"""

    azure_storage_account_key: Optional[TSecretStrValue] = None
    azure_storage_sas_token: TSecretStrValue = None
    azure_sas_token_permissions: str = "racwdl"
    """Permissions to use when generating a SAS token. Ignored when sas token is provided directly"""

    def to_adlfs_credentials(self) -> Dict[str, Any]:
        """Return a dict that can be passed as kwargs to adlfs"""
        return dict(
            account_name=self.azure_storage_account_name,
            account_key=self.azure_storage_account_key,
            sas_token=self.azure_storage_sas_token,
            account_host=self.azure_account_host,
        )

    def to_pyiceberg_fileio_config(self) -> Dict[str, Any]:
        return {
            "adls.account-name": self.azure_storage_account_name,
            "adls.account-key": self.azure_storage_account_key,
            "adls.sas-token": self.azure_storage_sas_token,
        }

    @classmethod
    def from_pyiceberg_fileio_config(cls, file_io: Dict[str, Any]) -> Self:
        # we'll modify file_io so make a copy
        file_io = copy(file_io)
        # convert signed uri to credentials
        for key, value in list(file_io.items()):
            if key.startswith("adls.sas-token."):
                if "adls.account-name" not in file_io:
                    file_io["adls.account-name"] = key.split(".")[2]
                if "adls.sas-token" not in file_io:
                    file_io["adls.sas-token"] = value  # key value is a sas token
        credentials: Self = cls()
        credentials.azure_account_host = file_io.get("adls.connection-string")
        credentials.azure_storage_account_key = file_io.get("adls.account-key")
        credentials.azure_storage_account_name = file_io.get("adls.account-name")
        credentials.azure_storage_sas_token = file_io.get("adls.sas-token")
        # if not credentials.is_partial():
        #     credentials.resolve()
        return credentials

    def create_sas_token(self) -> None:
        try:
            from azure.storage.blob import generate_account_sas, ResourceTypes
        except ModuleNotFoundError:
            raise MissingDependencyException(self.__class__.__name__, [_AZURE_STORAGE_EXTRA])

        self.azure_storage_sas_token = generate_account_sas(
            account_name=self.azure_storage_account_name,
            account_key=self.azure_storage_account_key,
            resource_types=ResourceTypes(container=True, object=True),
            permission=self.azure_sas_token_permissions,
            expiry=pendulum.now().add(days=1),
        )

    def on_partial(self) -> None:
        # sas token can be generated from account key
        if self.azure_storage_account_key and not self.azure_storage_sas_token:
            self.create_sas_token()
        if not self.is_partial():
            self.resolve()


@configspec
class AzureServicePrincipalCredentialsWithoutDefaults(AzureCredentialsBase, WithPyicebergConfig):
    azure_tenant_id: str = None
    azure_client_id: str = None
    azure_client_secret: TSecretStrValue = None

    def to_adlfs_credentials(self) -> Dict[str, Any]:
        return dict(
            account_name=self.azure_storage_account_name,
            account_host=self.azure_account_host,
            tenant_id=self.azure_tenant_id,
            client_id=self.azure_client_id,
            client_secret=self.azure_client_secret,
        )

    def to_pyiceberg_fileio_config(self) -> Dict[str, str]:
        return {
            "adls.account-name": self.azure_storage_account_name,
            "adls.tenant-id": self.azure_tenant_id,
            "adls.client-id": self.azure_client_id,
            "adls.client-secret": self.azure_client_secret,
        }

    @classmethod
    def from_pyiceberg_fileio_config(cls, file_io: Dict[str, Any]) -> Self:
        credentials: Self = cls()
        credentials.azure_tenant_id = file_io.get("adls.tenant-id")
        credentials.azure_client_id = file_io.get("adls.client-id")
        credentials.azure_storage_account_name = file_io.get("adls.account-name")
        credentials.azure_client_secret = file_io.get("adls.client-secret")
        # if not credentials.is_partial():
        #     credentials.resolve()
        return credentials


@configspec
class AzureCredentials(AzureCredentialsWithoutDefaults, CredentialsWithDefault):
    def on_partial(self) -> None:
        try:
            from azure.identity import DefaultAzureCredential
        except ModuleNotFoundError:
            raise MissingDependencyException(self.__class__.__name__, [_AZURE_STORAGE_EXTRA])

        if not self.azure_storage_account_key and not self.azure_storage_sas_token:
            self._set_default_credentials(DefaultAzureCredential())
            if self.azure_storage_account_name:
                self.resolve()
        else:
            super().on_partial()

    def to_adlfs_credentials(self) -> Dict[str, Any]:
        base_kwargs = super().to_adlfs_credentials()
        if self.has_default_credentials():
            base_kwargs["anon"] = False
        return base_kwargs


@configspec
class AzureServicePrincipalCredentials(
    AzureServicePrincipalCredentialsWithoutDefaults, CredentialsWithDefault
):
    def on_partial(self) -> None:
        try:
            from azure.identity import DefaultAzureCredential
        except ModuleNotFoundError:
            raise MissingDependencyException(self.__class__.__name__, [_AZURE_STORAGE_EXTRA])

        self._set_default_credentials(DefaultAzureCredential())
        if self.azure_storage_account_name:
            self.resolve()

    def to_adlfs_credentials(self) -> Dict[str, Any]:
        base_kwargs = super().to_adlfs_credentials()
        if self.has_default_credentials():
            base_kwargs["anon"] = False
        return base_kwargs


AnyAzureCredentials = Union[
    # Credentials without defaults come first because union types are attempted in order
    # and explicit config should supersede system defaults
    AzureCredentialsWithoutDefaults,
    AzureServicePrincipalCredentialsWithoutDefaults,
    AzureCredentials,
    AzureServicePrincipalCredentials,
]



import re
from functools import lru_cache

from dlt.common.normalizers.naming.snake_case import NamingConvention as SnakeCaseNamingConvention


class NamingConvention(SnakeCaseNamingConvention):
    """Case sensitive naming convention preserving all unicode characters except new line(s). Uses __ for path
    separation and will replace multiple underscores with a single one.
    """

    _CLEANUP_TABLE = str.maketrans('\n\r"', "___")

    @property
    def is_case_sensitive(self) -> bool:
        return True

    @staticmethod
    @lru_cache(maxsize=None)
    def _normalize_identifier(identifier: str, max_length: int) -> str:
        """Normalizes the identifier according to naming convention represented by this function"""

        normalized_ident = identifier.translate(NamingConvention._CLEANUP_TABLE)

        # shorten identifier
        return NamingConvention.shorten_identifier(
            NamingConvention.RE_UNDERSCORES.sub("_", normalized_ident), identifier, max_length
        )



import atexit
import time
import contextlib
import inspect
from typing import Any, Callable

from dlt.common.configuration.specs import RuntimeConfiguration
from dlt.common.exceptions import MissingDependencyException
from dlt.common.typing import TFun
from dlt.common.configuration import resolve_configuration
from dlt.common.runtime.anon_tracker import (
    TEventCategory,
    init_anon_tracker,
    disable_anon_tracker,
    track,
)

_TELEMETRY_STARTED = False


def start_telemetry(config: RuntimeConfiguration) -> None:
    # enable telemetry only once
    if is_telemetry_started():
        return

    if config.sentry_dsn:
        # may raise if sentry is not installed
        from dlt.common.runtime.sentry import init_sentry

        init_sentry(config)

    if config.dlthub_telemetry:
        init_anon_tracker(config)

    if config.dlthub_dsn:
        # TODO: we need pluggable modules for tracing so import into
        # concrete modules is not needed
        from dlt.pipeline.platform import init_platform_tracker

        init_platform_tracker()

    global _TELEMETRY_STARTED
    _TELEMETRY_STARTED = True


@atexit.register
def stop_telemetry() -> None:
    if not is_telemetry_started():
        return

    try:
        from dlt.common.runtime.sentry import disable_sentry

        disable_sentry()
    except (MissingDependencyException, ImportError):
        pass

    disable_anon_tracker()

    from dlt.pipeline.platform import disable_platform_tracker

    disable_platform_tracker()

    global _TELEMETRY_STARTED
    _TELEMETRY_STARTED = False


def is_telemetry_started() -> bool:
    return _TELEMETRY_STARTED


def with_telemetry(
    category: TEventCategory, command: str, track_before: bool, *args: str, **kwargs: Any
) -> Callable[[TFun], TFun]:
    """Decorator factory that attaches telemetry to a callable.

    The returned decorator wraps a function so that an anonymous telemetry event is
    emitted either before execution (if ``track_before`` is True) or after execution.
    When tracked after execution, the event includes whether the call succeeded.
    Telemetry is initialized lazily on first use if it is not already running.

    Args:
        category: telemetry event category used by the anon tracker.
        command: event/command name to report.
        track_before: if True, emit a single event before calling the function.
            if False, emit a single event after the call, including success state.
        *args: names of parameters from the decorated function whose values should
            be included in the event properties. names must match the function
            signature and can refer to positional or keyword parameters.
        **kwargs: additional key-value pairs to include in the event properties.

    Returns:
        a decorator that takes a function and returns a wrapped function with
        telemetry tracking applied. the wrapped function preserves the original
        signature and return value.

    Notes:
        - success is determined as follows when tracking after execution:
          - if the wrapped function returns an int, 0 is treated as success; any
            other value is treated as failure.
          - for non-int returns, success is True unless an exception is raised.
        - on exception, a failure event is emitted and the exception is re-raised.
        - event properties always include elapsed execution time in seconds under
          the 'elapsed' key and the success flag under 'success'.

    """

    def decorator(f: TFun) -> TFun:
        sig: inspect.Signature = inspect.signature(f)

        def _wrap(*f_args: Any, **f_kwargs: Any) -> Any:
            # look for additional arguments in call arguments
            bound_args = sig.bind(*f_args, **f_kwargs)
            props = {p: bound_args.arguments[p] for p in args if p in bound_args.arguments}
            # append additional props from kwargs
            props.update(kwargs)
            start_ts = time.monotonic()

            def _track(success: bool) -> None:
                with contextlib.suppress(Exception):
                    props["elapsed"] = time.monotonic() - start_ts
                    props["success"] = success
                    # resolve runtime config and init telemetry
                    if not _TELEMETRY_STARTED:
                        c = resolve_configuration(RuntimeConfiguration())
                        start_telemetry(c)
                    track(category, command, props)

            # some commands should be tracked before execution
            if track_before:
                _track(True)
                return f(*f_args, **f_kwargs)
            # some commands we track after, where we can pass the success
            try:
                rv = f(*f_args, **f_kwargs)
                # if decorated function returns int, 0 is a success - used to track dlt commands
                if isinstance(rv, int):
                    success = rv == 0
                else:
                    success = True
                _track(success)
                return rv
            except Exception:
                _track(False)
                raise

        return _wrap  # type: ignore

    return decorator



import argparse
import os
import sys
from typing import Optional

from dlt._workspace.cli import echo as fmt, utils
from dlt._workspace.cli import SupportsCliCommand, DEFAULT_VERIFIED_SOURCES_REPO
from dlt._workspace.cli.exceptions import CliCommandException
from dlt._workspace.cli.utils import add_mcp_arg_parser, make_mcp_run_flags
from dlt._workspace.cli._urls import (
    DLT_INIT_DOCS_URL,
    DLT_PIPELINE_COMMAND_DOCS_URL,
    DLT_TELEMETRY_DOCS_URL,
    DLT_DEPLOY_DOCS_URL,
)
from dlt.common.storages.configuration import TSchemaFileFormat

# NOTE: do not add command specific import here - do that inline to reduce import time


class InitCommand(SupportsCliCommand):
    command = "init"
    help_string = (
        "Creates a pipeline project in the current folder by adding existing verified source or"
        " creating a new one from template."
    )
    docs_url = DLT_INIT_DOCS_URL
    description = """
The `dlt init` command creates a new dlt pipeline script that loads data from `source` to `destination`. When you run the command, several things happen:

1. Creates a basic project structure if the current folder is empty by adding `.dlt/config.toml`, `.dlt/secrets.toml`, and `.gitignore` files.
2. Checks if the `source` argument matches one of our verified sources and, if so, adds it to your project.
3. If the `source` is unknown, uses a generic template to get you started.
4. Rewrites the pipeline scripts to use your `destination`.
5. Creates sample config and credentials in `secrets.toml` and `config.toml` for the specified source and destination.
6. Creates `requirements.txt` with dependencies required by the source and destination. If one exists, prints instructions on what to add to it.

This command can be used several times in the same folder to add more sources, destinations, and pipelines. It will also update the verified source code to the newest
version if run again with an existing `source` name. You will be warned if files will be overwritten or if the `dlt` version needs an upgrade to run a particular pipeline.
    """

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        self.parser = parser

        parser.add_argument(
            "--list-sources",
            "-l",
            default=False,
            action="store_true",
            help=(
                "Shows all available verified sources and their short descriptions. For each"
                " source, it checks if your local `dlt` version requires an update and prints the"
                " relevant warning."
            ),
        )
        parser.add_argument(
            "--list-destinations",
            default=False,
            action="store_true",
            help="Shows the name of all core dlt destinations.",
        )
        parser.add_argument(
            "source",
            nargs="?",
            help=(
                "Name of data source for which to create a pipeline. Adds existing verified"
                " source or creates a new pipeline template if verified source for your data"
                " source is not yet implemented."
            ),
        )
        parser.add_argument(
            "destination", nargs="?", help="Name of a destination ie. bigquery or redshift"
        )
        parser.add_argument(
            "--location",
            default=DEFAULT_VERIFIED_SOURCES_REPO,
            help="Advanced. Uses a specific url or local path to verified sources repository.",
        )
        parser.add_argument(
            "--branch",
            default=None,
            help=(
                "Advanced. Uses specific branch of the verified sources repository to fetch the"
                " template."
            ),
        )

        parser.add_argument(
            "--eject",
            default=False,
            action="store_true",
            help=(
                "Ejects the source code of the core source like sql_database or rest_api so they"
                " will be editable by you."
            ),
        )

    def execute(self, args: argparse.Namespace) -> None:
        from dlt._workspace.cli._init_command import (
            list_destinations_command_wrapper,
            list_sources_command_wrapper,
            init_command_wrapper,
        )

        if args.list_sources:
            list_sources_command_wrapper(args.location, args.branch)
        elif args.list_destinations:
            list_destinations_command_wrapper()
        else:
            if not args.source or not args.destination:
                self.parser.print_usage()
                raise CliCommandException()
            else:
                init_command_wrapper(
                    args.source,
                    args.destination,
                    args.location,
                    args.branch,
                    args.eject,
                )


class PipelineCommand(SupportsCliCommand):
    command = "pipeline"
    help_string = "Operations on pipelines that were ran locally"
    docs_url = DLT_PIPELINE_COMMAND_DOCS_URL
    description = """
The `dlt pipeline` command provides a set of commands to inspect the pipeline working directory, tables, and data in the destination and check for problems encountered during data loading.
    """

    def configure_parser(self, pipe_cmd: argparse.ArgumentParser) -> None:
        # keep here to avoid importing schema storages at cli startup
        from dlt.common.storages.configuration import SCHEMA_FILES_EXTENSIONS

        self.parser = pipe_cmd

        pipe_cmd.add_argument(
            "--list-pipelines",
            "-l",
            default=False,
            action="store_true",
            help="List local pipelines",
        )
        pipe_cmd.add_argument("pipeline_name", nargs="?", help="Pipeline name")
        pipe_cmd.add_argument("--pipelines-dir", help="Pipelines working directory", default=None)
        pipe_cmd.add_argument(
            "--verbose",
            "-v",
            action="count",
            default=0,
            help="Provides more information for certain commands.",
            dest="verbosity",
        )

        pipeline_subparsers = pipe_cmd.add_subparsers(
            title="Available subcommands", dest="operation", required=False
        )

        pipe_cmd_sync_parent = argparse.ArgumentParser(add_help=False)
        pipe_cmd_sync_parent.add_argument(
            "--destination", help="Sync from this destination when local pipeline state is missing."
        )
        pipe_cmd_sync_parent.add_argument(
            "--dataset-name", help="Dataset name to sync from when local pipeline state is missing."
        )

        pipeline_subparsers.add_parser(
            "info",
            help="Displays state of the pipeline, use -v or -vv for more info",
            description="""
Displays the content of the working directory of the pipeline: dataset name, destination, list of
schemas, resources in schemas, list of completed and normalized load packages, and optionally a
pipeline state set by the resources during the extraction process.
""",
        )
        show_cmd = pipeline_subparsers.add_parser(
            "show",
            help=(
                "Generates and launches workspace dashboard with the loading status and dataset"
                " explorer"
            ),
            description="""
Launches the workspace dashboard with a comprehensive interface to inspect the pipeline state, schemas, and data in the destination.

This dashboard should be executed from the same folder from which you ran the pipeline script to be able access destination credentials.

If the --edit flag is used, will launch the editable version of the dashboard if it exists in the current directory, or create this version and launch it in edit mode.

Requires `marimo` to be installed in the current environment: `pip install marimo`.
""",
        )
        show_cmd.add_argument(
            "--edit",
            default=False,
            action="store_true",
            help=(
                "Creates editable version of workspace dashboard in current directory if it does"
                " not exist there yet and launches it in edit mode."
            ),
        )
        pipeline_subparsers.add_parser(
            "failed-jobs",
            help=(
                "Displays information on all the failed loads in all completed packages, failed"
                " jobs and associated error messages"
            ),
            description="""
This command scans all the load packages looking for failed jobs and then displays information on
files that got loaded and the failure message from the destination.
""",
        )
        pipeline_subparsers.add_parser(
            "drop-pending-packages",
            help=(
                "Deletes all extracted and normalized packages including those that are partially"
                " loaded."
            ),
            description="""
Removes all extracted and normalized packages in the pipeline's working dir.
`dlt` keeps extracted and normalized load packages in the pipeline working directory. When the `run` method is called, it will attempt to normalize and load
pending packages first. The command above removes such packages. Note that **pipeline state** is not reverted to the state at which the deleted packages
were created. Using `dlt pipeline ... sync` is recommended if your destination supports state sync.
""",
        )
        pipeline_subparsers.add_parser(
            "sync",
            help=(
                "Drops the local state of the pipeline and resets all the schemas and restores it"
                " from destination. The destination state, data and schemas are left intact."
            ),
            description="""
This command will remove the pipeline working directory with all pending packages, not synchronized
state changes, and schemas and retrieve the last synchronized data from the destination. If you drop
the dataset the pipeline is loading to, this command results in a complete reset of the pipeline state.

In case of a pipeline without a working directory, the command may be used to create one from the
destination. In order to do that, you need to pass the dataset name and destination name to the CLI
and provide the credentials to connect to the destination (i.e., in `.dlt/secrets.toml`) placed in the
folder where you execute the `pipeline sync` command.
""",
            parents=[pipe_cmd_sync_parent],
        )
        pipeline_subparsers.add_parser(
            "trace",
            help="Displays last run trace, use -v or -vv for more info",
            description="""
Displays the trace of the last pipeline run containing the start date of the run, elapsed time, and the
same information for all the steps (`extract`, `normalize`, and `load`). If any of the steps failed,
you'll see the message of the exceptions that caused that problem. Successful `load` and `run` steps
will display the load info instead.
""",
        )
        pipe_cmd_schema = pipeline_subparsers.add_parser(
            "schema",
            help="Displays default schema",
            description="Displays the default schema for the selected pipeline.",
        )
        pipe_cmd_schema.add_argument(
            "--format",
            choices=SCHEMA_FILES_EXTENSIONS,
            default="yaml",
            help="Display schema in this format",
        )
        pipe_cmd_schema.add_argument(
            "--remove-defaults",
            action="store_true",
            help="Does not show default hint values",
            default=True,
        )

        pipe_cmd_drop = pipeline_subparsers.add_parser(
            "drop",
            help="Selectively drop tables and reset state",
            description="""
Selectively drop tables and reset state.

```sh
dlt pipeline <pipeline name> drop [resource_1] [resource_2]
```

Drops tables generated by selected resources and resets the state associated with them. Mainly used
to force a full refresh on selected tables. In the example below, we drop all tables generated by
the `repo_events` resource in the GitHub pipeline:

```sh
dlt pipeline github_events drop repo_events
```

`dlt` will inform you of the names of dropped tables and the resource state slots that will be
reset:

```text
About to drop the following data in dataset airflow_events_1 in destination dlt.destinations.duckdb:
Selected schema:: github_repo_events
Selected resource(s):: ['repo_events']
Table(s) to drop:: ['issues_event', 'fork_event', 'pull_request_event', 'pull_request_review_event', 'pull_request_review_comment_event', 'watch_event', 'issue_comment_event', 'push_event__payload__commits', 'push_event']
Resource(s) state to reset:: ['repo_events']
Source state path(s) to reset:: []
Do you want to apply these changes? [y/N]
```

As a result of the command above the following will happen:

1. All the indicated tables will be dropped in the destination. Note that `dlt` drops the nested
   tables as well.
2. All the indicated tables will be removed from the indicated schema.
3. The state for the resource `repo_events` was found and will be reset.
4. New schema and state will be stored in the destination.

The `drop` command accepts several advanced settings:

1. You can use regexes to select resources. Prepend the `re:` string to indicate a regex pattern. The example
   below will select all resources starting with `repo`:

```sh
dlt pipeline github_events drop "re:^repo"
```

2. You can drop all tables in the indicated schema:

```sh
dlt pipeline chess drop --drop-all
```

3. You can indicate additional state slots to reset by passing JsonPath to the source state. In the example
   below, we reset the `archives` slot in the source state:

```sh
dlt pipeline chess_pipeline drop --state-paths archives
```

This will select the `archives` key in the `chess` source.

```json
{
  "sources":{
    "chess": {
      "archives": [
        "https://api.chess.com/pub/player/magnuscarlsen/games/2022/05"
      ]
    }
  }
}
```

**This command is still experimental** and the interface will most probably change.

""",
            parents=[pipe_cmd_sync_parent],
            epilog=(
                f"See {DLT_PIPELINE_COMMAND_DOCS_URL}#selectively-drop-tables-and-reset-state for"
                " more info"
            ),
        )
        pipe_cmd_drop.add_argument(
            "resources",
            nargs="*",
            help=(
                "One or more resources to drop. Can be exact resource name(s) or regex pattern(s)."
                " Regex patterns must start with re:"
            ),
        )
        pipe_cmd_drop.add_argument(
            "--drop-all",
            action="store_true",
            default=False,
            help="Drop all resources found in schema. Supersedes [resources] argument.",
        )
        pipe_cmd_drop.add_argument(
            "--state-paths", nargs="*", help="State keys or json paths to drop", default=()
        )
        pipe_cmd_drop.add_argument(
            "--schema",
            help="Schema name to drop from (if other than default schema).",
            dest="schema_name",
        )
        pipe_cmd_drop.add_argument(
            "--state-only",
            action="store_true",
            help="Only wipe state for matching resources without dropping tables.",
            default=False,
        )

        pipe_cmd_package = pipeline_subparsers.add_parser(
            "load-package",
            help="Displays information on load package, use -v or -vv for more info",
            description="""
Shows information on a load package with a given `load_id`. The `load_id` parameter defaults to the
most recent package. Package information includes its state (`COMPLETED/PROCESSED`) and list of all
jobs in a package with their statuses, file sizes, types, and in case of failed jobs—the error
messages from the destination. With the verbose flag set `dlt pipeline -v ...`, you can also see the
list of all tables and columns created at the destination during the loading of that package.
""",
        )
        pipe_cmd_package.add_argument(
            "load_id",
            metavar="load-id",
            nargs="?",
            help="Load id of completed or normalized package. Defaults to the most recent package.",
        )

        DEFAULT_PIPELINE_MCP_PORT = 43656
        add_mcp_arg_parser(
            pipeline_subparsers,
            "This MCP facilitates schema and data exploration for the dataset created with this"
            " pipeline",
            "Launch MCP server attached to this pipeline",
            DEFAULT_PIPELINE_MCP_PORT,
        )

    def execute(self, args: argparse.Namespace) -> None:
        from dlt._workspace.cli._pipeline_command import pipeline_command_wrapper

        if (
            args.list_pipelines
            or args.operation == "list"
            or (not args.pipeline_name and not args.operation)
        ):
            # Always use max verbosity (1) for dlt pipeline list - show full details
            pipeline_command_wrapper("list", "-", args.pipelines_dir, 1)
        else:
            command_kwargs = dict(args._get_kwargs())
            if not command_kwargs.get("pipeline_name"):
                self.parser.print_usage()
                raise CliCommandException(error_code=-1)
            command_kwargs["operation"] = args.operation or "info"
            del command_kwargs["command"]
            del command_kwargs["list_pipelines"]
            pipeline_command_wrapper(**command_kwargs)


class SchemaCommand(SupportsCliCommand):
    command = "schema"
    help_string = "Shows, converts and upgrades schemas"
    docs_url = "https://dlthub.com/docs/reference/command-line-interface#dlt-schema"
    description = """
The `dlt schema` command will load, validate and print out a dlt schema: `dlt schema path/to/my_schema_file.yaml`.
    """

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        # keep here to avoid importing schema storages at cli startup
        from dlt.common.storages.configuration import SCHEMA_FILES_EXTENSIONS

        self.parser = parser

        parser.add_argument(
            "file",
            help="Schema file name, in yaml or json format, will autodetect based on extension",
        )
        parser.add_argument(
            "--format",
            choices=SCHEMA_FILES_EXTENSIONS,
            default="yaml",
            help="Display schema in this format",
        )
        parser.add_argument(
            "--remove-defaults",
            action="store_true",
            help="Does not show default hint values",
            default=True,
        )

    def execute(self, args: argparse.Namespace) -> None:
        # keep here to avoid importing schema/yaml at cli startup
        import yaml

        from dlt.common import json
        from dlt.common.schema.schema import Schema
        from dlt.common.typing import DictStrAny

        @utils.track_command("schema", False, "format_")
        def schema_command_wrapper(
            file_path: str, format_: TSchemaFileFormat, remove_defaults: bool
        ) -> None:
            with open(file_path, "rb") as f:
                if os.path.splitext(file_path)[1][1:] == "json":
                    schema_dict: DictStrAny = json.load(f)
                else:
                    schema_dict = yaml.safe_load(f)
            s = Schema.from_dict(schema_dict)
            export = utils.fetch_schema_export(s, format_=format_, remove_defaults=remove_defaults)
            fmt.echo(export["content"])

        schema_command_wrapper(args.file, args.format, args.remove_defaults)


class DashboardCommand(SupportsCliCommand):
    command = "dashboard"
    help_string = "Starts the dlt workspace dashboard"
    description = """
The `dlt dashboard` command starts the dlt workspace dashboard. You can use the dashboard:

* to list and inspect local pipelines
* browse the full pipeline schema and all hints
* browse the data in the destination
* inspect the pipeline state

    """

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        self.parser = parser
        self.parser.add_argument(
            "--pipelines-dir", help="Pipelines working directory", default=None
        )
        self.parser.add_argument(
            "--edit",
            action="store_true",
            help="Eject Dashboard and start editable version",
            default=None,
        )

    def execute(self, args: argparse.Namespace) -> None:
        @utils.track_command("dashboard", True)
        def dashboard_command_wrapper(pipelines_dir: Optional[str], edit: bool) -> None:
            from dlt._workspace.helpers.dashboard.runner import run_dashboard

            run_dashboard(pipelines_dir=pipelines_dir, edit=edit)

        dashboard_command_wrapper(pipelines_dir=args.pipelines_dir, edit=args.edit)


class TelemetryCommand(SupportsCliCommand):
    command = "telemetry"
    help_string = "Shows telemetry status"
    docs_url = DLT_TELEMETRY_DOCS_URL
    description = """
The `dlt telemetry` command shows the current status of dlt telemetry. Learn more about telemetry and what we send in our telemetry docs.
    """

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        self.parser = parser

    def execute(self, args: argparse.Namespace) -> None:
        from dlt._workspace.cli._telemetry_command import telemetry_status_command_wrapper

        telemetry_status_command_wrapper()


class DeployCommand(SupportsCliCommand):
    command = "deploy"
    help_string = "Creates a deployment package for a selected pipeline script"
    docs_url = DLT_DEPLOY_DOCS_URL
    description = """
The `dlt deploy` command prepares your pipeline for deployment and gives you step-by-step instructions on how to accomplish it. To enable this functionality, please first execute `pip install "dlt[cli]"` which will add additional packages to the current environment.
    """

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        # keep here to avoid importing heavy deploy deps at cli startup
        from dlt._workspace.cli._deploy_command import (
            DeploymentMethods,
            COMMAND_DEPLOY_REPO_LOCATION,
            SecretFormats,
        )

        self.parser = parser
        deploy_cmd = parser
        deploy_comm = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False
        )

        deploy_cmd.add_argument(
            "pipeline_script_path", metavar="pipeline-script-path", help="Path to a pipeline script"
        )

        deploy_comm.add_argument(
            "--location",
            default=COMMAND_DEPLOY_REPO_LOCATION,
            help="Advanced. Uses a specific url or local path to pipelines repository.",
        )
        deploy_comm.add_argument(
            "--branch",
            help="Advanced. Uses specific branch of the deploy repository to fetch the template.",
        )

        deploy_sub_parsers = deploy_cmd.add_subparsers(
            title="Available subcommands", dest="deployment_method"
        )

        # deploy github actions
        deploy_github_cmd = deploy_sub_parsers.add_parser(
            DeploymentMethods.github_actions.value,
            help="Deploys the pipeline to Github Actions",
            parents=[deploy_comm],
            description="""
Deploys the pipeline to GitHub Actions.

GitHub Actions (https://github.com/features/actions) is a CI/CD runner with a large free tier which you can use to run your pipelines.

You must specify when the GitHub Action should run using a cron schedule expression. The command also takes additional flags:
`--run-on-push` (default is False) and `--run-manually` (default is True). Remember to put the cron
schedule expression in quotation marks.

For the chess.com API example from our docs, you can deploy it with `dlt deploy chess.py github-action --schedule "*/30 * * * *"`.

Follow the guide on how to deploy a pipeline with GitHub Actions in our documentation for more information.
""",
        )
        deploy_github_cmd.add_argument(
            "--schedule",
            required=True,
            help=(
                "A schedule with which to run the pipeline, in cron format. Example: '*/30 * * * *'"
                " will run the pipeline every 30 minutes. Remember to enclose the scheduler"
                " expression in quotation marks!"
            ),
        )
        deploy_github_cmd.add_argument(
            "--run-manually",
            default=True,
            action="store_true",
            help="Allows the pipeline to be run manually form Github Actions UI.",
        )
        deploy_github_cmd.add_argument(
            "--run-on-push",
            default=False,
            action="store_true",
            help="Runs the pipeline with every push to the repository.",
        )

        # deploy airflow composer
        deploy_airflow_cmd = deploy_sub_parsers.add_parser(
            DeploymentMethods.airflow_composer.value,
            help="Deploys the pipeline to Airflow",
            parents=[deploy_comm],
            description="""
Google Composer (https://cloud.google.com/composer?hl=en) is a managed Airflow environment provided by Google. Follow the guide in our docs on how to deploy a pipeline with Airflow to learn more. This command will:


* create an Airflow DAG for your pipeline script that you can customize. The DAG uses
the `dlt` Airflow wrapper (https://github.com/dlt-hub/dlt/blob/devel/dlt/helpers/airflow_helper.py#L37) to make this process trivial.

* provide you with the environment variables and secrets that you must add to Airflow.

* provide you with a cloudbuild file to sync your GitHub repository with the `dag` folder of your Airflow Composer instance.
""",
        )
        deploy_airflow_cmd.add_argument(
            "--secrets-format",
            default=SecretFormats.toml.value,
            choices=[v.value for v in SecretFormats],
            required=False,
            help="Format of the secrets",
        )

    def execute(self, args: argparse.Namespace) -> None:
        # keep here: pipdeptree scans all installed packages on import
        try:
            import pipdeptree  # noqa: F401
            import cron_descriptor  # noqa: F401

            deploy_command_available = True
        except ImportError:
            deploy_command_available = False
        except Exception:
            # beartype import hook can break pip._vendor.distlib on Windows;
            # pipdeptree is installed but cannot be imported — let deploy proceed,
            # generate_pip_freeze will handle the failure gracefully
            deploy_command_available = True

        # exit if deploy command is not available
        if not deploy_command_available:
            fmt.warning(
                "Please install additional command line dependencies to use deploy command:"
            )
            fmt.secho('pip install "dlt[cli]"', bold=True)
            fmt.echo(
                "We ask you to install those dependencies separately to keep our core library small"
                " and make it work everywhere."
            )
            raise CliCommandException()

        deploy_args = vars(args)
        if deploy_args.get("deployment_method") is None:
            self.parser.print_help()
            raise CliCommandException()
        else:
            from dlt._workspace.cli._deploy_command import deploy_command_wrapper

            deploy_command_wrapper(
                pipeline_script_path=deploy_args.pop("pipeline_script_path"),
                deployment_method=deploy_args.pop("deployment_method"),
                repo_location=deploy_args.pop("location"),
                branch=deploy_args.pop("branch"),
                **deploy_args,
            )


# NOTE: we should port this as a command or as a tool to the core repo
class CliDocsCommand(SupportsCliCommand):
    command = "render-docs"
    help_string = "Renders markdown version of cli docs"
    description = """
The `dlt render-docs` command renders markdown version of cli docs by parsing the argparse help output and generating a markdown file.
If you are reading this on the docs website, you are looking at the rendered version of the cli docs generated by this command.
    """

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        self.parser = parser

        self.parser.add_argument("file_name", nargs=1, help="Output file name")

        self.parser.add_argument(
            "--commands",
            nargs="*",
            help="List of command names to render (optional)",
            default=None,
        )

        self.parser.add_argument(
            "--compare",
            default=False,
            action="store_true",
            help="Compare the changes and raise if output would be updated",
        )

    def execute(self, args: argparse.Namespace) -> None:
        from dlt._workspace.cli._dlt import _create_parser
        from dlt._workspace.cli._docs_command import render_argparse_markdown

        parser, _ = _create_parser()

        result = render_argparse_markdown("dlt", parser, commands=args.commands)

        if args.compare:
            with open(args.file_name[0], "r", encoding="utf-8") as f:
                if result != f.read():
                    fmt.error(
                        "Cli Docs out of date, please update, please run "
                        "update-cli-docs from the main Makefile and commit your changes. "
                    )
                    raise CliCommandException()

            fmt.note("Docs page up to date.")
        else:
            with open(args.file_name[0], "w", encoding="utf-8") as f:
                f.write(result)
            fmt.note("Docs page updated.")


class AiCommand(SupportsCliCommand):
    command = "ai"
    help_string = "Use AI-powered development tools and utilities"
    # docs_url =
    description = (
        "The `dlt ai` command provides commands to configure your LLM-enabled IDE and MCP server."
    )

    def configure_parser(self, ai_cmd: argparse.ArgumentParser) -> None:
        self.parser = ai_cmd

        ai_subparsers = ai_cmd.add_subparsers(
            title="Available subcommands", dest="operation", required=False
        )

        # status command
        ai_subparsers.add_parser(
            "status",
            help="Show AI setup status: dlt version, agent, toolkits, readiness checks",
        )

        # init command
        from dlt._workspace.cli._urls import (
            DEFAULT_AI_WORKBENCH_BRANCH,
            DEFAULT_AI_WORKBENCH_REPO,
        )

        init_cmd = ai_subparsers.add_parser(
            "init",
            help="Install initial AI rules and skills for your AI coding agent",
        )
        init_cmd.add_argument(
            "--agent",
            choices=["claude", "cursor", "codex"],
            default=None,
            help="AI coding agent to install for. Auto-detected if omitted.",
        )
        init_cmd.add_argument(
            "--location",
            default=DEFAULT_AI_WORKBENCH_REPO,
            help="Advanced. Git URL or local path to AI workbench repository.",
        )
        init_cmd.add_argument(
            "--branch",
            default=DEFAULT_AI_WORKBENCH_BRANCH,
            help="Advanced. Git branch to fetch from.",
        )
        init_cmd.add_argument(
            "--overwrite",
            default=False,
            action="store_true",
            help="Overwrite existing files instead of skipping them.",
        )

        # secrets command group
        secrets_cmd = ai_subparsers.add_parser(
            "secrets",
            help="Manage secrets files used by dlt",
            description="List, view (redacted), or update secret files used by dlt providers.",
        )
        secrets_subparsers = secrets_cmd.add_subparsers(
            title="Secrets subcommands", dest="secrets_operation", required=False
        )
        secrets_subparsers.add_parser(
            "list",
            help="List secret file locations from providers",
        )
        view_cmd = secrets_subparsers.add_parser(
            "view-redacted",
            help="Print secrets TOML with all values replaced by '***'",
            description=(
                "Without --path, shows the unified view merged from all project"
                " secret files. With --path, shows that exact file."
            ),
        )
        view_cmd.add_argument(
            "--path",
            default=None,
            help="Show this exact file instead of the unified provider view",
        )
        update_cmd = secrets_subparsers.add_parser(
            "update-fragment",
            help="Merge a TOML fragment into the secrets file",
        )
        update_cmd.add_argument(
            "fragment",
            nargs="?",
            default=None,
            help="TOML fragment string to merge; reads from stdin if omitted",
        )
        update_cmd.add_argument(
            "--path",
            required=True,
            help="Path to the secrets TOML file to write to",
        )

        # toolkit command group
        toolkit_cmd = ai_subparsers.add_parser(
            "toolkit",
            help="Manage AI toolkit plugins (list, info, install)",
        )
        toolkit_cmd.add_argument(
            "name", nargs="?", help="Toolkit name (required for info and install)"
        )
        toolkit_sub = toolkit_cmd.add_subparsers(dest="toolkit_operation", required=False)

        # shared parent with --location and --branch
        toolkit_common = argparse.ArgumentParser(add_help=False)
        toolkit_common.add_argument(
            "--location",
            default=DEFAULT_AI_WORKBENCH_REPO,
            help="Advanced. Git URL or local path to toolkit repository.",
        )
        toolkit_common.add_argument(
            "--branch",
            default=DEFAULT_AI_WORKBENCH_BRANCH,
            help="Advanced. Git branch to fetch toolkit from.",
        )

        toolkit_sub.add_parser(
            "list",
            help="List available toolkits",
            parents=[toolkit_common],
        )
        toolkit_sub.add_parser(
            "info",
            help="Show toolkit contents and components",
            parents=[toolkit_common],
        )
        install_cmd = toolkit_sub.add_parser(
            "install",
            help="Install toolkit components into project",
            parents=[toolkit_common],
        )
        install_cmd.add_argument(
            "--agent",
            choices=["claude", "cursor", "codex"],
            default=None,
            help="AI coding agent to install for. Auto-detected if omitted.",
        )
        install_cmd.add_argument(
            "--overwrite",
            default=False,
            action="store_true",
            help="Overwrite existing files instead of skipping them.",
        )
        install_cmd.add_argument(
            "--strict",
            default=False,
            action="store_true",
            help="Fail on validation warnings (invalid frontmatter, etc.).",
        )

        # shared run flags — used by both `dlt ai mcp [flags]` and `dlt ai mcp run [flags]`
        mcp_run_flags = make_mcp_run_flags(default_port=8000)

        mcp_cmd = ai_subparsers.add_parser(
            "mcp",
            help="Run or install the dlt MCP server",
            parents=[mcp_run_flags],
        )
        mcp_sub = mcp_cmd.add_subparsers(dest="mcp_operation", required=False)

        mcp_sub.add_parser("run", help="Start the MCP server (default)", parents=[mcp_run_flags])

        mcp_install_cmd = mcp_sub.add_parser(
            "install",
            help="Install MCP server config into the current project",
        )
        mcp_install_cmd.add_argument(
            "--agent",
            choices=["claude", "cursor", "codex"],
            default=None,
            help="AI coding agent to install for. Auto-detected if omitted.",
        )
        mcp_install_cmd.add_argument(
            "--features",
            nargs="*",
            default=None,
            help="MCP feature sets to include in the server config",
        )
        mcp_install_cmd.add_argument(
            "--name",
            default="dlt-workspace",
            help="Server name in the MCP config (default: dlt-workspace)",
        )
        mcp_install_cmd.add_argument(
            "--overwrite",
            default=False,
            action="store_true",
            help="Overwrite existing server config instead of skipping.",
        )

    def execute(self, args: argparse.Namespace) -> None:
        from dlt._workspace.cli._urls import (
            DEFAULT_AI_WORKBENCH_BRANCH,
            DEFAULT_AI_WORKBENCH_REPO,
        )
        from dlt._workspace.cli.ai import (
            ai_status_command,
            ai_init_command,
            ai_mcp_run_command,
            ai_mcp_install_command,
            ai_secrets_list_command,
            ai_secrets_view_redacted_command,
            ai_secrets_update_fragment_command,
            ai_toolkit_install_command,
            ai_toolkit_list_command,
            ai_toolkit_info_command,
        )

        if args.operation == "status":
            ai_status_command()
        elif args.operation == "init":
            ai_init_command(
                agent=args.agent,
                location=args.location,
                branch=args.branch,
                overwrite=args.overwrite,
            )
        elif args.operation == "secrets":
            op = getattr(args, "secrets_operation", None)
            if op == "view-redacted":
                ai_secrets_view_redacted_command(path=args.path)
            elif op == "update-fragment":
                fragment = args.fragment or sys.stdin.read()
                ai_secrets_update_fragment_command(fragment=fragment, path=args.path)
            else:
                ai_secrets_list_command()
        elif args.operation == "toolkit":
            tk_op = getattr(args, "toolkit_operation", None)
            if tk_op == "list":
                ai_toolkit_list_command(
                    location=args.location,
                    branch=args.branch,
                )
            elif tk_op == "info":
                if not args.name:
                    fmt.error("Toolkit name is required for 'info'.")
                    raise CliCommandException()
                ai_toolkit_info_command(
                    name=args.name,
                    location=args.location,
                    branch=args.branch,
                )
            elif tk_op == "install":
                if not args.name:
                    fmt.error("Toolkit name is required for 'install'.")
                    raise CliCommandException()
                ai_toolkit_install_command(
                    name=args.name,
                    agent=args.agent,
                    location=args.location,
                    branch=args.branch,
                    overwrite=args.overwrite,
                    strict=args.strict,
                )
            else:
                # default: list toolkits
                ai_toolkit_list_command(
                    location=getattr(args, "location", DEFAULT_AI_WORKBENCH_REPO),
                    branch=getattr(args, "branch", DEFAULT_AI_WORKBENCH_BRANCH),
                )
        elif args.operation == "mcp":
            mcp_op = getattr(args, "mcp_operation", None)
            if mcp_op == "install":
                ai_mcp_install_command(
                    agent=args.agent,
                    features=args.features,
                    name=args.name,
                    overwrite=args.overwrite,
                )
            else:
                # default: run
                ai_mcp_run_command(
                    port=getattr(args, "port", 8000),
                    stdio=getattr(args, "stdio", False),
                    sse=getattr(args, "sse", False),
                    features=getattr(args, "features", None),
                )
        else:
            self.parser.print_usage()


class WorkspaceCommand(SupportsCliCommand):
    command = "workspace"
    help_string = "Manage current Workspace"
    description = """
Commands to get info, cleanup local files and launch Workspace MCP. Run without command get
workspace info.
"""

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        self.parser = parser

        parser.add_argument(
            "--verbose",
            "-v",
            action="count",
            default=0,
            help="Provides more information for certain commands.",
            dest="verbosity",
        )

        subparsers = parser.add_subparsers(
            title="Available subcommands", dest="workspace_command", required=False
        )

        # clean command
        clean_local_parser = subparsers.add_parser(
            "clean",
            help=(
                "Cleans local data for the selected profile. Locally loaded data will be deleted. "
                "Pipelines working directories are also deleted by default. Data in remote "
                "destinations is not affected."
            ),
        )
        clean_local_parser.add_argument(
            "--skip-data-dir",
            action="store_true",
            default=False,
            help="Do not delete pipelines working dir.",
        )

        subparsers.add_parser(
            "info",
            help="Displays workspace info.",
        )

        DEFAULT_DLT_MCP_PORT = 43654
        add_mcp_arg_parser(
            subparsers,
            "This MCP allows to attach to any pipeline that was previously ran in this workspace"
            " and then facilitates schema and data exploration in the pipeline's dataset.",
            "Launch dlt MCP server in current Python environment and Workspace",
            DEFAULT_DLT_MCP_PORT,
        )

        show_parser = subparsers.add_parser(
            "show",
            help="Shows Workspace Dashboard for the pipelines and data in this workspace.",
        )

        show_parser.add_argument(
            "--edit",
            action="store_true",
            help="Eject Dashboard and start editable version",
            default=None,
        )

    def execute(self, args: argparse.Namespace) -> None:
        from dlt._workspace._workspace_context import active
        from dlt._workspace.cli._workspace_command import (
            print_workspace_info,
            clean_workspace,
            show_workspace,
            start_mcp,
        )

        workspace_context = active()

        if args.workspace_command == "info" or not args.workspace_command:
            print_workspace_info(workspace_context, args.verbosity)
        elif args.workspace_command == "clean":
            clean_workspace(workspace_context, args)
        elif args.workspace_command == "show":
            show_workspace(workspace_context, args.edit)
        elif args.workspace_command == "mcp":
            start_mcp(workspace_context, port=args.port, stdio=args.stdio, sse=args.sse)
        else:
            self.parser.print_usage()


class ProfileCommand(SupportsCliCommand):
    command = "profile"
    help_string = "Manage Workspace built-in profiles"
    description = """
Commands to list and pin profiles
Run without arguments to list all profiles, the default profile and the
pinned profile in current project.
"""

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        self.parser = parser

        parser.add_argument("profile_name", help="Name of the profile", nargs="?")

        subparsers = parser.add_subparsers(
            title="Available subcommands", dest="profile_command", required=False
        )

        subparsers.add_parser(
            "info",
            help="Show information about the current profile.",
            description="Show information about the current profile.",
        )

        subparsers.add_parser(
            "list",
            help="Show list of built-in profiles.",
            description="Show list of built-in profiles.",
        )

        subparsers.add_parser(
            "pin",
            help="Pin a profile to the Workspace.",
            description="""
Pin a profile to the Workspace, this will be the new default profile while it is pinned.
""",
        )

    def execute(self, args: argparse.Namespace) -> None:
        from dlt._workspace._workspace_context import active
        from dlt._workspace.cli._profile_command import (
            print_profile_info,
            list_profiles,
            pin_profile,
        )

        workspace_context = active()

        if args.profile_command == "info" or not args.profile_command:
            print_profile_info(workspace_context)
        elif args.profile_command == "list":
            list_profiles(workspace_context)
        elif args.profile_command == "pin":
            pin_profile(workspace_context, args.profile_name)
        else:
            self.parser.print_usage()



import os
import glob
import semver
from typing import ClassVar, Sequence

from semver import Version

from dlt.common.configuration import with_config, known_sections
from dlt.common.configuration.accessors import config
from dlt.common.storages.exceptions import StorageMigrationError
from dlt.common.storages.versioned_storage import VersionedStorage
from dlt.common.storages.file_storage import FileStorage
from dlt.common.storages.load_package import PackageStorage
from dlt.common.storages.configuration import NormalizeStorageConfiguration
from dlt.common.utils import set_working_dir


class NormalizeStorage(VersionedStorage):
    STORAGE_VERSION: ClassVar[str] = "1.0.1"
    EXTRACTED_FOLDER: ClassVar[str] = (
        "extracted"  # folder within the volume where extracted files to be normalized are stored
    )

    @with_config(spec=NormalizeStorageConfiguration, sections=(known_sections.NORMALIZE,))
    def __init__(
        self, is_owner: bool, config: NormalizeStorageConfiguration = config.value
    ) -> None:
        super().__init__(
            NormalizeStorage.STORAGE_VERSION,
            is_owner,
            FileStorage(config.normalize_volume_path, "t", makedirs=is_owner),
        )
        self.config = config
        if is_owner:
            self.initialize_storage()
        self.extracted_packages = PackageStorage(
            FileStorage(os.path.join(self.storage.storage_path, NormalizeStorage.EXTRACTED_FOLDER)),
            "extracted",
        )

    def initialize_storage(self) -> None:
        self.storage.create_folder(NormalizeStorage.EXTRACTED_FOLDER, exists_ok=True)

    def list_files_to_normalize_sorted(self) -> Sequence[str]:
        """Gets all data files in extracted packages storage. This method is compatible with current and all past storages"""
        root_dir = os.path.join(self.storage.storage_path, NormalizeStorage.EXTRACTED_FOLDER)
        with set_working_dir(root_dir):
            files = glob.glob("**/*", recursive=True)
            # return all files that are not schema files
            return sorted(
                [
                    file
                    for file in files
                    if not file.endswith(PackageStorage.SCHEMA_FILE_NAME)
                    and os.path.isfile(file)
                    and not file.endswith(PackageStorage.LOAD_PACKAGE_STATE_FILE_NAME)
                ]
            )

    def migrate_storage(self, from_version: Version, to_version: Version) -> None:
        if from_version == "1.0.0" and from_version < to_version:
            # get files in storage
            if len(self.list_files_to_normalize_sorted()) > 0:
                raise StorageMigrationError(
                    self.storage.storage_path,
                    from_version,
                    to_version,
                    f"There are extract files in folder `{NormalizeStorage.EXTRACTED_FOLDER}`."
                    " Storage will not migrate automatically due to possible data loss. Delete the"
                    " files or normalize it with dlt 0.3.x",
                )
            from_version = semver.Version.parse("1.0.1")
            self._save_version(from_version)



"""Generic API Source"""

from copy import deepcopy
from typing import Any, Dict, List, Optional, cast, Union
import graphlib
from requests.auth import AuthBase

import dlt
from dlt.common.validation import validate_dict
from dlt.common import jsonpath
from dlt.common.schema.schema import Schema
from dlt.common.schema.typing import TSchemaContract
from dlt.common.utils import exclude_keys
from dlt.common.exceptions import ValueErrorWithKnownValues

from dlt.extract import Incremental, DltResource, DltSource, decorators

from dlt.sources.helpers.rest_client import RESTClient
from dlt.sources.helpers.rest_client.paginators import BasePaginator
from dlt.sources.helpers.rest_client.auth import (
    HttpBasicAuth,
    BearerTokenAuth,
    APIKeyAuth,
    AuthConfigBase,
    OAuth2ClientCredentials,
)
from dlt.sources.helpers.rest_client.typing import HTTPMethodBasic
from dlt.sources.helpers.rest_client.redaction import SENSITIVE_PARAMS
from .typing import (
    AuthConfig,
    ClientConfig,
    EndpointResourceBase,
    ResolvedParam,
    ResolveParamConfig,
    Endpoint,
    EndpointResource,
    IncrementalParamConfig,
    RESTAPIConfig,
    ParamBindType,
    ProcessingSteps,
)
from .config_setup import (
    IncrementalParam,
    create_auth,
    create_paginator,
    build_resource_dependency_graph,
    paginate_dependent_resource,
    paginate_resource,
    process_parent_data_item,
    setup_incremental_object,
    create_response_hooks,
    expand_placeholders,
    convert_incremental_values,
)
from .utils import check_connection  # noqa: F401

PARAM_TYPES: List[ParamBindType] = ["incremental", "resolve"]
MIN_SECRET_MASKING_LENGTH = 3
SENSITIVE_KEYS: List[str] = list(SENSITIVE_PARAMS) + ["username", "client_id"]


@decorators.source
def rest_api(
    client: ClientConfig = dlt.config.value,
    resources: List[Union[str, EndpointResource, DltResource]] = dlt.config.value,
    resource_defaults: Optional[EndpointResourceBase] = None,
) -> List[DltResource]:
    """Creates and configures a REST API source with default settings"""
    return rest_api_resources(
        {
            "client": client,
            "resources": resources,
            "resource_defaults": resource_defaults,
        }
    )


def rest_api_source(
    config: RESTAPIConfig,
    name: str = None,
    section: str = None,
    max_table_nesting: int = None,
    root_key: bool = None,
    schema: Schema = None,
    schema_contract: TSchemaContract = None,
    parallelized: bool = False,
) -> DltSource:
    """Creates and configures a REST API source for data extraction.

    Args:
        config (RESTAPIConfig): Configuration for the REST API source.
        name (str, optional): Name of the source.
        section (str, optional): Section of the configuration file.
        max_table_nesting (int, optional): Maximum depth of nested table above which
            the remaining nodes are loaded as structs or JSON.
        root_key (bool, optional): Enables merging on all resources by propagating
            root foreign key to child tables. This option is most useful if you
            plan to change write disposition of a resource to disable/enable merge.
            Defaults to False.
        schema (Schema, optional): An explicit `Schema` instance to be associated
            with the source. If not present, `dlt` creates a new `Schema` object
            with provided `name`. If such `Schema` already exists in the same
            folder as the module containing the decorated function, such schema
            will be loaded from file.
        schema_contract (TSchemaContract, optional): Schema contract settings
            that will be applied to this resource.
        parallelized (bool, optional): If `True`, resource generators will be
            extracted in parallel with other resources. Transformers that return items are also parallelized.
            Non-eligible resources are ignored. Defaults to `False` which preserves resource settings.

    Returns:
        DltSource: A configured dlt source.

    Example:
        pokemon_source = rest_api_source({
            "client": {
                "base_url": "https://pokeapi.co/api/v2/",
                "paginator": "json_link",
            },
            "resources": [
                {
                    "name": "pokemon",
                    "endpoint": {
                        "path": "pokemon",
                        "params": {
                            "limit": 100,
                        },
                    },
                    "primary_key": "id",
                }
            ]
        })
    """
    # TODO: this must be removed when TypedDicts are supported by resolve_configuration
    #   so secrets values are bound BEFORE validation. validation will happen during the resolve process
    _validate_config(config)
    decorated = rest_api.clone(
        name=name,
        section=section,
        max_table_nesting=max_table_nesting,
        root_key=root_key,
        schema=schema,
        schema_contract=schema_contract,
        parallelized=parallelized,
    )

    return decorated(**config)


def rest_api_resources(config: RESTAPIConfig) -> List[DltResource]:
    """Creates a list of resources from a REST API configuration.

    Args:
        config (RESTAPIConfig): Configuration for the REST API source.

    Returns:
        List[DltResource]: List of dlt resources.

    Example:
        github_source = rest_api_resources({
            "client": {
                "base_url": "https://api.github.com/repos/dlt-hub/dlt/",
                "auth": {
                    "token": dlt.secrets["token"],
                },
            },
            "resource_defaults": {
                "primary_key": "id",
                "write_disposition": "merge",
                "endpoint": {
                    "params": {
                        "per_page": 100,
                    },
                },
            },
            "resources": [
                {
                    "name": "issues",
                    "endpoint": {
                        "path": "issues",
                        "params": {
                            "sort": "updated",
                            "direction": "desc",
                            "state": "open",
                            "since": "{incremental.start_value}",
                        },
                        "incremental": {
                            "cursor_path": "updated_at",
                            "initial_value": "2024-01-25T11:21:28Z",
                        },
                    },
                },
                {
                    "name": "issue_comments",
                    "endpoint": {
                        "path": "issues/{resources.issues.number}/comments",
                    },
                },
            ],
        })
    """

    _validate_config(config)

    client_config = config["client"]
    resource_defaults = config.get("resource_defaults") or {}
    resource_list = config["resources"]

    (
        dependency_graph,
        endpoint_resource_map,
        resolved_param_map,
    ) = build_resource_dependency_graph(
        resource_defaults,
        resource_list,
    )

    resources = create_resources(
        client_config,
        dependency_graph,
        endpoint_resource_map,
        resolved_param_map,
    )

    return list(resources.values())


def create_resources(
    client_config: ClientConfig,
    dependency_graph: graphlib.TopologicalSorter,  # type: ignore[type-arg]
    endpoint_resource_map: Dict[str, Union[EndpointResource, DltResource]],
    resolved_param_map: Dict[str, Optional[List[ResolvedParam]]],
) -> Dict[str, DltResource]:
    resources = {}

    for resource_name in dependency_graph.static_order():
        resource_name = cast(str, resource_name)
        endpoint_resource = endpoint_resource_map[resource_name]
        if isinstance(endpoint_resource, DltResource):
            resources[resource_name] = endpoint_resource
            continue

        endpoint_config = cast(Endpoint, endpoint_resource["endpoint"])
        request_params = endpoint_config.get("params", {})
        request_json = endpoint_config.get("json")
        request_data = endpoint_config.get("data")
        request_headers = endpoint_config.get("headers")
        paginator = create_paginator(endpoint_config.get("paginator"))
        processing_steps = endpoint_resource.pop("processing_steps", [])

        resolved_params: List[ResolvedParam] = resolved_param_map[resource_name]

        include_from_parent: List[str] = endpoint_resource.get("include_from_parent", [])
        if not resolved_params and include_from_parent:
            raise ValueError(
                f"Resource `{resource_name}` has `include_from_parent` but is not "
                "dependent on another resource"
            )
        _validate_param_type(request_params)
        (
            incremental_object,
            incremental_param,
            incremental_cursor_transform,
        ) = setup_incremental_object(request_params, endpoint_config.get("incremental"))

        client = RESTClient(
            base_url=client_config["base_url"],
            headers=client_config.get("headers"),
            auth=create_auth(endpoint_config.get("auth", client_config.get("auth"))),
            paginator=create_paginator(client_config.get("paginator")),
            session=client_config.get("session"),
        )

        hooks = create_response_hooks(endpoint_config.get("response_actions"))

        resource_kwargs = exclude_keys(endpoint_resource, {"endpoint", "include_from_parent"})

        def process(
            resource: DltResource,
            processing_steps: List[ProcessingSteps],
        ) -> Any:
            for step in processing_steps:
                if "filter" in step:
                    resource.add_filter(step["filter"])
                if "map" in step:
                    resource.add_map(step["map"])
                if "yield_map" in step:
                    resource.add_yield_map(step["yield_map"])
            return resource

        if resolved_params is None:
            resources[resource_name] = dlt.resource(
                paginate_resource,
                **resource_kwargs,  # TODO: implement typing.Unpack
            )(
                method=endpoint_config.get("method", "GET"),
                path=endpoint_config.get("path"),
                headers=request_headers,
                params=request_params,
                json=request_json,
                data=request_data,
                paginator=paginator,
                data_selector=endpoint_config.get("data_selector"),
                hooks=hooks,
                client=client,
                incremental_object=incremental_object,
                incremental_param=incremental_param,
                incremental_cursor_transform=incremental_cursor_transform,
            )

            resources[resource_name] = process(resources[resource_name], processing_steps)

        else:
            first_param = resolved_params[0]
            predecessor = resources[first_param.resolve_config["resource"]]

            base_params = exclude_keys(request_params, {x.param_name for x in resolved_params})

            parallelized = resource_kwargs.pop("parallelized", False)

            resources[resource_name] = dlt.resource(  # type: ignore[call-overload]
                paginate_dependent_resource,
                data_from=predecessor,
                **resource_kwargs,  # TODO: implement typing.Unpack
            )(
                method=endpoint_config.get("method", "get"),
                path=endpoint_config.get("path"),
                headers=request_headers,
                params=base_params,
                json=request_json,
                data=request_data,
                paginator=paginator,
                data_selector=endpoint_config.get("data_selector"),
                hooks=hooks,
                client=client,
                resolved_params=resolved_params,
                include_from_parent=include_from_parent,
                incremental_object=incremental_object,
                incremental_param=incremental_param,
                incremental_cursor_transform=incremental_cursor_transform,
                parallelized=parallelized,
            )

            resources[resource_name] = process(resources[resource_name], processing_steps)

    return resources


def _validate_config(config: RESTAPIConfig) -> None:
    c = deepcopy(config)
    client_config = c.get("client")
    if client_config:
        auth = client_config.get("auth")
        if auth:
            _mask_secrets(auth)
    resources = c.get("resources", [])
    for resource in resources:
        if isinstance(resource, (str, DltResource)):
            continue
        if endpoint := resource.get("endpoint"):
            if not isinstance(endpoint, str):
                auth = endpoint.get("auth")
                if auth:
                    _mask_secrets(auth)

    validate_dict(RESTAPIConfig, c, path=".")


def _mask_secrets(auth_config: AuthConfig) -> AuthConfig:
    # skip AuthBase (derived from requests lib) or shorthand notation
    if (
        isinstance(auth_config, AuthBase)
        and not isinstance(auth_config, AuthConfigBase)
        or isinstance(auth_config, str)
    ):
        return auth_config

    has_sensitive_key = any(key in auth_config for key in SENSITIVE_KEYS)
    if (
        isinstance(
            auth_config,
            (APIKeyAuth, BearerTokenAuth, HttpBasicAuth, OAuth2ClientCredentials),
        )
        or has_sensitive_key
    ):
        return _mask_secrets_dict(auth_config)
    # Here, we assume that OAuth2 and other custom classes that don't implement __get__()
    # also don't print secrets in __str__()
    # TODO: call auth_config.mask_secrets() when that is implemented
    return auth_config


def _mask_secrets_dict(auth_config: AuthConfig) -> AuthConfig:
    for sensitive_key in SENSITIVE_KEYS:
        try:
            auth_config[sensitive_key] = _mask_secret(auth_config[sensitive_key])  # type: ignore[literal-required, index]
        except KeyError:
            continue
    return auth_config


def _mask_secret(secret: Optional[str]) -> str:
    if secret is None:
        return "None"
    if len(secret) < MIN_SECRET_MASKING_LENGTH:
        return "*****"
    return f"{secret[0]}*****{secret[-1]}"


def _validate_param_type(
    request_params: Dict[str, Union[ResolveParamConfig, IncrementalParamConfig, Any]],
) -> None:
    for param_name, value in request_params.items():
        if isinstance(value, dict) and value.get("type") not in PARAM_TYPES:
            raise ValueErrorWithKnownValues(f"{param_name}['type']", value.get("type"), PARAM_TYPES)



import abc
import contextlib
from typing import Any, Dict, Optional, Set, Tuple

from dlt.common import logger
from dlt.common.typing import AnyType
from dlt.common.pendulum import pendulum
from dlt.common.configuration.specs import known_sections
from dlt.common.configuration.specs.base_configuration import is_secret_hint

from .doc import BaseDocProvider

SECRETS_TOML_KEY = "dlt_secrets_toml"


class VaultDocProvider(BaseDocProvider):
    """A toml-backed Vault abstract config provider.

    This provider allows implementation of providers that store secrets in external vaults: like Hashicorp, Google Secrets or Airflow Metadata.
    The basic working principle is obtain config and secrets values from Vault keys and reconstitute a `secrets.toml` like document that is then used
    as a cache.

    The implemented must provide `_look_vault` method that returns a value from external vault from external key.

    To reduce number of calls to external vaults the provider is searching for a known configuration fragments which should be toml documents and merging
    them with the
    - only keys with secret type hint (CredentialsConfiguration, TSecretValue) will be looked up by default.
    - provider gathers `toml` document fragments that contain source and destination credentials in path specified below
    - single values will not be retrieved, only toml fragments by default

    """

    def __init__(
        self, only_secrets: bool, only_toml_fragments: bool, list_secrets: bool = False
    ) -> None:
        """Initializes the toml backed Vault provider by loading a toml fragment from `dlt_secrets_toml` key and using it as initial configuration.

        Args:
            only_secrets (bool): Only looks for secret values (CredentialsConfiguration, TSecretValue) by returning None (not found)
            only_toml_fragments (bool): Only load the known toml fragments and ignore any other lookups by returning None (not found)
            list_secrets (bool): When True, lists all available secrets once on first access to avoid unnecessary lookups
        """
        self.only_secrets = only_secrets
        self.only_toml_fragments = only_toml_fragments
        self.list_secrets = list_secrets
        self._vault_lookups: Dict[str, Any] = {}
        self._available_keys: Optional[Set[str]] = None
        if list_secrets and (only_toml_fragments or only_secrets):
            logger.warning(
                "Listing secrets is enabled so available keys are known upfront but one or both"
                f" only_toml_fragments ({only_toml_fragments}) only_secrets ({only_secrets}) are"
                " enabled too  so lookups of certain keys may be still skipped. Disable both to"
                " access all keys in vault."
            )
        super().__init__({})

    def get_value(
        self, key: str, hint: type, pipeline_name: str, *sections: str
    ) -> Tuple[Optional[Any], str]:
        # global settings must be updated first
        self._update_from_vault(SECRETS_TOML_KEY, None, AnyType, None, ())
        # then regular keys
        full_key = self.get_key_name(key, pipeline_name, *sections)
        value, _ = super().get_value(key, hint, pipeline_name, *sections)
        if value is None:
            if self.only_secrets and not is_secret_hint(hint):
                pass
            else:
                # only secrets hints are handled fully
                self._load_fragments(key, pipeline_name, *sections)
                value, _ = super().get_value(key, hint, pipeline_name, *sections)
                # skip checking the exact path if we check only toml fragments
                if value is None and not self.only_toml_fragments:
                    # look for key in the vault and update the toml document
                    self._update_from_vault(full_key, key, hint, pipeline_name, sections)
                    value, _ = super().get_value(key, hint, pipeline_name, *sections)

        return value, full_key

    @property
    def supports_secrets(self) -> bool:
        return True

    def clear_lookup_cache(self) -> None:
        self._vault_lookups.clear()

    def _load_fragments(self, key: str, pipeline_name: str, *sections: str) -> None:
        """Load known toml fragments from the vault
        * pipeline scoped SECRETS TOML KEY
        * destination and sources key
        * destination.<name> and sources.<name> key
        * the above scoped to pipeline
        """
        if pipeline_name:
            # loads dlt_secrets_toml for particular pipeline
            lookup_fk = self.get_key_name(SECRETS_TOML_KEY, pipeline_name)
            self._update_from_vault(lookup_fk, "", AnyType, pipeline_name, ())

        # generate auxiliary paths to get from vault
        for known_section in [known_sections.SOURCES, known_sections.DESTINATION]:

            def _look_at_idx(idx: int, full_path: Tuple[str, ...], pipeline_name: str) -> None:
                lookup_key = full_path[idx]
                lookup_sections = full_path[:idx]
                lookup_fk = self.get_key_name(lookup_key, *lookup_sections)
                self._update_from_vault(
                    lookup_fk, lookup_key, AnyType, pipeline_name, lookup_sections
                )

            def _lookup_paths(pipeline_name_: str, known_section_: str) -> None:
                with contextlib.suppress(ValueError):
                    full_path = sections + (key,)
                    if pipeline_name_:
                        full_path = (pipeline_name_,) + full_path
                    idx = full_path.index(known_section_)
                    _look_at_idx(idx, full_path, pipeline_name_)
                    # if there's element after index then also try it (destination name / source name)
                    if len(full_path) - 1 > idx:
                        _look_at_idx(idx + 1, full_path, pipeline_name_)

            # first query the shortest paths so the longer paths can override it
            _lookup_paths(None, known_section)  # check sources and sources.<source_name>
            if pipeline_name:
                _lookup_paths(
                    pipeline_name, known_section
                )  # check <pipeline_name>.sources and <pipeline_name>.sources.<source_name>

    @abc.abstractmethod
    def _look_vault(self, full_key: str, hint: type) -> str:
        """Looks for `full_key` in the vault, may use `hint` to detect if expected value is a secret"""
        pass

    @abc.abstractmethod
    def _list_vault(self) -> Set[str]:
        """Lists keys in the vault in order to skip lookups for non existing keys

        Returns:
            Set[str]: A set of available key names in the vault
        """
        pass

    def _update_from_vault(
        self, full_key: str, key: str, hint: type, pipeline_name: str, sections: Tuple[str, ...]
    ) -> None:
        if full_key in self._vault_lookups:
            return

        # list the vault on first access if enabled
        if self.list_secrets and self._available_keys is None:
            self._available_keys = self._list_vault()

        # skip lookup if we know the key doesn't exist
        if (
            self.list_secrets
            and self._available_keys is not None
            and full_key not in self._available_keys
        ):
            self._vault_lookups[full_key] = pendulum.now()
            return

        # print(f"tries '{key}' {pipeline_name} | {sections} at '{full_key}'")
        logger.debug(f"Vault provider {self.name} will make a request for {full_key}")
        secret = self._look_vault(full_key, hint)
        self._vault_lookups[full_key] = pendulum.now()
        if secret is not None:
            self.set_fragment(key, secret, pipeline_name, *sections)

    @property
    def is_empty(self) -> bool:
        return False



from abc import ABC, abstractmethod
from typing import (
    Awaitable,
    Callable,
    Dict,
    List,
    Any,
    Generic,
    Tuple,
    Union,
    overload,
    ClassVar,
    Type,
)

from dlt.common import logger
from dlt.common.configuration.specs import BaseConfiguration, known_sections

from dlt.common.reflection.ref import object_from_ref
from dlt.common.runtime.run_context import get_plugin_modules
from dlt.common.schema import Schema
from dlt.common.schema.typing import TSchemaContract
from dlt.common.typing import ParamSpec, Self, TypeVar
from dlt.common.warnings import Dlt04DeprecationWarning, deprecated

from dlt.extract.source import DltSource
from dlt.extract.exceptions import UnknownSourceReference

TDltSourceImpl = TypeVar("TDltSourceImpl", bound=DltSource, default=DltSource)
TSourceFunParams = ParamSpec("TSourceFunParams")


class SourceFactory(ABC, Generic[TSourceFunParams, TDltSourceImpl]):
    name: str
    """Name of the source which is also the name of the associated schema."""
    section: str
    """Configuration section right after 'sources' in default layout."""
    max_table_nesting: int
    """Maximum depth of nested table above which remaining nodes are loaded as structs or JSON."""
    root_key: bool
    """Enables merging on all resources by propagating root foreign key to nested tables."""
    schema: Schema
    """An explicit `Schema` instance to be associated with the source."""
    schema_contract: TSchemaContract
    """Schema contract settings that will be applied to this source."""
    spec: Type[BaseConfiguration]
    """A specification of configuration and secret values required by the source."""
    parallelized: bool
    """If `True`, resource generators will be extracted in parallel with other resources."""

    # TODO: model after Destination, which also needs to be broken down into reference and factory
    ref: "SourceReference"

    def __call__(
        self, *args: TSourceFunParams.args, **kwargs: TSourceFunParams.kwargs
    ) -> TDltSourceImpl:
        pass

    @abstractmethod
    def add_postprocessor(
        self, func: Callable[[TDltSourceImpl], Union[TDltSourceImpl, Awaitable[TDltSourceImpl]]]
    ) -> None:
        """Adds a callback that receives and returns a DltSource after it is created."""

    @abstractmethod
    def clone(
        self,
        *,
        name: str = None,
        section: str = None,
        max_table_nesting: int = None,
        root_key: bool = None,
        schema: Schema = None,
        schema_contract: TSchemaContract = None,
        spec: Type[BaseConfiguration] = None,
        parallelized: bool = None,
        _impl_cls: Type[TDltSourceImpl] = DltSource,  # type: ignore[assignment]
    ) -> Self:
        """Overrides default decorator arguments that will be used to when DltSource instance and returns modified clone."""

    with_args = deprecated("Please use clone method instead", category=Dlt04DeprecationWarning)(
        clone
    )


AnySourceFactory = SourceFactory[Any, DltSource]


class SourceReference:
    SOURCES: ClassVar[Dict[str, "SourceReference"]] = {}
    """A registry of all the decorated sources and resources discovered when importing modules"""

    ref: str
    """A fully qualified reference: __module__.__name__ to the factory instance"""
    SPEC: Type[BaseConfiguration]
    factory: AnySourceFactory
    section: str
    name: str

    def __init__(
        self,
        ref: str,
        SPEC: Type[BaseConfiguration],
        factory: AnySourceFactory,
        section: str,
        name: str,
    ) -> None:
        self.ref = ref
        self.SPEC = SPEC
        self.factory = factory
        self.section = section
        self.name = name

    @classmethod
    def expand_shorthand_ref(cls, ref: str) -> List[str]:
        """Converts ref into fully qualified form, return one or more alternatives for shorthand notations.
        Run context is injected if needed. Following formats are recognized
        - section.name
        - name
        """
        ref_split = ref.split(".")
        ref_parts = len(ref_split)
        if ref_parts < 3:
            # expand with known prefixes
            refs = []
            for ref_prefix in get_plugin_modules():
                if ref_prefix:
                    ref_prefix = f"{ref_prefix}.{known_sections.SOURCES}"
                else:
                    ref_prefix = f"{known_sections.SOURCES}"
                # expand shorthand notation
                if ref_parts == 1:
                    refs.append(f"{ref_prefix}.{ref}.{ref}")
                elif ref_parts == 2:
                    # for ref with two parts two options are possible
                    refs.append(f"{ref_prefix}.{ref}")
            return refs
        return []

    @classmethod
    def register(cls, ref_obj: "SourceReference") -> None:
        ref = ref_obj.ref
        if ref in cls.SOURCES:
            logger.debug(f"A source with ref {ref} is already registered and will be overwritten")
        cls.SOURCES[ref] = ref_obj

    @overload
    @classmethod
    def find(cls, ref: str) -> AnySourceFactory: ...

    @overload
    @classmethod
    def find(
        cls,
        ref: str,
        /,
        raise_exec_errors: bool = False,
        import_missing_modules: bool = False,
        _impl_sig: None = ...,
        _impl_cls: Type[TDltSourceImpl] = None,
    ) -> SourceFactory[Any, TDltSourceImpl]: ...

    @overload
    @classmethod
    def find(
        cls,
        ref: str,
        /,
        raise_exec_errors: bool = False,
        import_missing_modules: bool = False,
        _impl_sig: Callable[TSourceFunParams, Any] = None,
        _impl_cls: Type[TDltSourceImpl] = None,
    ) -> SourceFactory[TSourceFunParams, TDltSourceImpl]: ...

    @classmethod
    def find(
        cls,
        ref: str,
        raise_exec_errors: bool = False,
        import_missing_modules: bool = False,
        _impl_sig: Callable[TSourceFunParams, Any] = None,
        _impl_cls: Type[TDltSourceImpl] = None,
    ) -> Any:
        """Returns source factory from reference `ref`. Looks into registry or tries auto-import.
        Expands shorthand notation into section.name eg. "sql_database" is expanded into
            "dlt.sources.sql_database.sql_database".

        You can control auto-import behavior:
        - `raise_exec_errors` - will re-raise code execution errors in imported modules
        - `import_missing_modules` - will ignore missing dependencies during import by substituting
           them with dummy modules. this should be only used to manipulate local dev environment
        """
        refs = cls.expand_shorthand_ref(ref)
        if ref not in refs:
            refs = [ref] + refs

        for ref_ in refs:
            if wrapper := cls.SOURCES.get(ref_):
                return wrapper.factory

        import_traces = []

        # try to import module
        for possible_type in refs:
            if "." not in possible_type:
                continue
            factory, trace = object_from_ref(
                possible_type,
                SourceReference._factory_typechecker,
                raise_exec_errors=raise_exec_errors,
                import_missing_modules=import_missing_modules,
            )
            if factory:
                return factory
            import_traces.append(trace)

        raise UnknownSourceReference(ref, refs, traces=import_traces)

    @classmethod
    def from_reference(
        cls,
        ref: str,
        /,
        name: str = None,
        section: str = None,
        max_table_nesting: int = None,
        root_key: bool = None,
        schema: Schema = None,
        schema_contract: TSchemaContract = None,
        spec: Type[BaseConfiguration] = None,
        parallelized: bool = None,
        _impl_cls: Type[TDltSourceImpl] = None,
        source_args: Tuple[Any, ...] = None,
        source_kwargs: Dict[str, Any] = None,
    ) -> TDltSourceImpl:
        """Find registered source factory or imports it, then instantiates the DltSource using
        passed args and kwargs.
        Passes additional arguments to `clone` of source factory
        """
        source_args = source_args or ()
        source_kwargs = source_kwargs or {}
        return cls.find(ref, _impl_cls=_impl_cls).clone(
            name=name,
            section=section,
            max_table_nesting=max_table_nesting,
            root_key=root_key,
            schema=schema,
            schema_contract=schema_contract,
            spec=spec,
            parallelized=parallelized,
            _impl_cls=_impl_cls,
        )(*source_args, **source_kwargs)

    @staticmethod
    def _factory_typechecker(factory: Any) -> Any:
        if hasattr(factory, "_factory"):
            factory = factory._factory
            # make sure it is factory interface (we could check Protocol as well)
        if not isinstance(factory, SourceFactory):
            raise TypeError("expected AnySourceFactory type")
        return factory



import os

from dlt.common.configuration.specs.pluggable_run_context import RunContextBase

DEFAULT_PROFILE = "dev"
PIN_FILE_NAME = "profile-name"

BUILT_IN_PROFILES = {
    "dev": "dev profile, workspace default",
    "prod": "production profile, assumed by pipelines deployed in Runtime",
    "tests": "profile assumed when running tests",
    "access": (
        "production profile, assumed by interactive notebooks in Runtime, typically with limited"
        " access rights"
    ),
}
LOCAL_PROFILES = ["dev", "tests"]
"""Profiles used in local workspace"""


def get_profile_pin_file(context: RunContextBase) -> str:
    return context.get_setting(PIN_FILE_NAME)


def read_profile_pin(context: RunContextBase) -> str:
    # load profile name from .profile-name
    profile_pin = get_profile_pin_file(context)
    if os.path.isfile(profile_pin):
        with open(profile_pin, "tr", encoding="utf-8") as pin:
            return pin.readline().strip()
    return None


def save_profile_pin(context: RunContextBase, profile: str) -> None:
    # load profile name from .profile-name
    profile_pin = get_profile_pin_file(context)
    with open(profile_pin, "wt", encoding="utf-8") as pin:
        pin.write(profile)
